import numpy as np
import pytest
from numpy.testing import assert_allclose

from stcal.ramp_fitting.ols_cas22 import JUMP_DET, Parameter, Variance, fit_ramps
from stcal.ramp_fitting.ols_cas22._jump import (
    FixedOffsets,
    PixelOffsets,
    _fill_pixel_values,
    fill_fixed_values,
)
from stcal.ramp_fitting.ols_cas22._ramp import from_read_pattern, init_ramps

# Purposefully set a fixed seed so that the tests in this module are deterministic
RNG = np.random.default_rng(619)

# The read time is constant for the given telescope/instrument so we set it here
#    to be the one for Roman as it is known to be a reasonable value
READ_TIME = 3.04

# Choose small read noise relative to the flux to make it extremely unlikely
#    that the random process will "accidentally" generate a set of data, which
#    can trigger jump detection. This makes it easier to cleanly test jump
#    detection is doing what we expect.
FLUX = 100
READ_NOISE = np.float32(5)

# Set a value for jumps which makes them obvious relative to the normal flux
JUMP_VALUE = 1_000

# Choose reasonable values for arbitrary test parameters, these are kept the same
#    across all tests to make it easier to isolate the effects of something using
#    multiple tests.
N_PIXELS = 100_000
CHI2_TOL = 0.03
GOOD_PROB = 0.7


def test_init_ramps():
    """
    Test turning dq flags into initial ramp splits
        Note that because `init_ramps` itself returns a stack, which does not have
        a direct python equivalent, we call the wrapper for `init_ramps` which
        converts that stack into a list ordered in the same fashion as the stack
    """
    # from stcal.ramp_fitting.ols_cas22._core import _init_ramps_list

    dq = np.array(
        [
            [0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1],
            [0, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1, 1],
            [0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 1],
            [0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1],
        ],
        dtype=np.int32,
    )

    n_resultants, n_pixels = dq.shape
    ramps = [init_ramps(dq[:, index], n_resultants) for index in range(n_pixels)]

    assert len(ramps) == dq.shape[1] == 16

    # Check that the ramps are correct

    # No DQ
    assert ramps[0] == [{"start": 0, "end": 3}]

    # 1 DQ
    assert ramps[1] == [{"start": 1, "end": 3}]
    assert ramps[2] == [{"start": 0, "end": 0}, {"start": 2, "end": 3}]
    assert ramps[3] == [{"start": 0, "end": 1}, {"start": 3, "end": 3}]
    assert ramps[4] == [{"start": 0, "end": 2}]

    # 2 DQ
    assert ramps[5] == [{"start": 2, "end": 3}]
    assert ramps[6] == [{"start": 1, "end": 1}, {"start": 3, "end": 3}]
    assert ramps[7] == [{"start": 1, "end": 2}]
    assert ramps[8] == [{"start": 0, "end": 0}, {"start": 3, "end": 3}]
    assert ramps[9] == [{"start": 0, "end": 0}, {"start": 2, "end": 2}]
    assert ramps[10] == [{"start": 0, "end": 1}]

    # 3 DQ
    assert ramps[11] == [{"start": 3, "end": 3}]
    assert ramps[12] == [{"start": 2, "end": 2}]
    assert ramps[13] == [{"start": 1, "end": 1}]
    assert ramps[14] == [{"start": 0, "end": 0}]

    # 4 DQ
    assert ramps[15] == []


@pytest.fixture(scope="module")
def read_pattern():
    """
    Basic data for simulating ramps for testing (not unpacked)

    Returns
    -------
        read_pattern : list[list[int]]
            The example read pattern
        metadata : dict
            The metadata computed from the read pattern
    """
    return [
        [1, 2, 3, 4],
        [5],
        [6, 7, 8],
        [9, 10, 11, 12, 13, 14, 15, 16, 17, 18],
        [19, 20, 21],
        [22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36],
    ]


def test_from_read_pattern(read_pattern):
    """Test turning read_pattern into the time data"""
    metadata = from_read_pattern(read_pattern, READ_TIME, len(read_pattern))._to_dict()  # noqa: SLF001

    t_bar = metadata["t_bar"]
    tau = metadata["tau"]
    n_reads = metadata["n_reads"]

    # Check that the data is correct
    assert_allclose(t_bar, [7.6, 15.2, 21.279999, 41.040001, 60.799999, 88.159996])
    assert_allclose(tau, [5.7, 15.2, 19.928888, 36.023998, 59.448887, 80.593781])
    assert np.all(n_reads == [4, 1, 3, 10, 3, 15])

    # Check datatypes
    assert t_bar.dtype == np.float32
    assert tau.dtype == np.float32
    assert n_reads.dtype == np.int32


@pytest.fixture(scope="module")
def ramp_data(read_pattern):
    """
    Basic data for simulating ramps for testing (not unpacked)

    Returns
    -------
        read_pattern : list[list[int]]
            The example read pattern
        metadata : dict
            The metadata computed from the read pattern
    """
    data = from_read_pattern(read_pattern, READ_TIME, len(read_pattern))._to_dict()  # noqa: SLF001

    return data["t_bar"], data["tau"], data["n_reads"], read_pattern


def test_fill_fixed_values(ramp_data):
    """Test computing the fixed data for all pixels"""
    t_bar, tau, n_reads, _ = ramp_data

    n_resultants = len(t_bar)
    fixed = np.empty((FixedOffsets.n_fixed_offsets, n_resultants - 1), dtype=np.float32)
    fixed = fill_fixed_values(fixed, t_bar, tau, n_reads, n_resultants)

    # Sanity check that the shape of fixed is correct
    assert fixed.shape == (2 * 4, n_resultants - 1)

    # Split into the different types of data
    t_bar_diffs = fixed[FixedOffsets.single_t_bar_diff : FixedOffsets.double_t_bar_diff + 1, :]
    t_bar_diff_sqrs = fixed[FixedOffsets.single_t_bar_diff_sqr : FixedOffsets.double_t_bar_diff_sqr + 1, :]
    read_recip = fixed[FixedOffsets.single_read_recip : FixedOffsets.double_read_recip + 1, :]
    var_slope_vals = fixed[FixedOffsets.single_var_slope_val : FixedOffsets.double_var_slope_val + 1, :]

    # Sanity check that these are all the right shape
    assert t_bar_diffs.shape == (2, n_resultants - 1)
    assert t_bar_diff_sqrs.shape == (2, n_resultants - 1)
    assert read_recip.shape == (2, n_resultants - 1)
    assert var_slope_vals.shape == (2, n_resultants - 1)

    # Check the computed data
    #   These are computed using loop in cython, here we check against numpy
    # Single diffs
    assert np.all(t_bar_diffs[0] == t_bar[1:] - t_bar[:-1])
    assert np.all(t_bar_diff_sqrs[0] == (t_bar[1:] - t_bar[:-1]) ** 2)
    assert np.all(read_recip[0] == np.float32(1 / n_reads[1:]) + np.float32(1 / n_reads[:-1]))
    assert np.all(var_slope_vals[0] == (tau[1:] + tau[:-1] - 2 * np.minimum(t_bar[1:], t_bar[:-1])))

    # Double diffs
    assert np.all(t_bar_diffs[1, :-1] == t_bar[2:] - t_bar[:-2])
    assert np.all(t_bar_diff_sqrs[1, :-1] == (t_bar[2:] - t_bar[:-2]) ** 2)
    assert np.all(read_recip[1, :-1] == np.float32(1 / n_reads[2:]) + np.float32(1 / n_reads[:-2]))
    assert np.all(var_slope_vals[1, :-1] == (tau[2:] + tau[:-2] - 2 * np.minimum(t_bar[2:], t_bar[:-2])))

    # Last double diff should be NaN
    assert np.isnan(t_bar_diffs[1, -1])
    assert np.isnan(t_bar_diff_sqrs[1, -1])
    assert np.isnan(read_recip[1, -1])
    assert np.isnan(var_slope_vals[1, -1])


def _generate_resultants(read_pattern, n_pixels=1):
    """
    Generate a set of resultants for a pixel

    Parameters:
        read_pattern : list[list[int]]
            The read pattern to use
        n_pixels:
            The number of pixels to generate resultants for. Default is 1.

    Returns:
        resultants
            The resultants generated
    """
    resultants = np.zeros((len(read_pattern), n_pixels), dtype=np.float32)

    # Use Poisson process to simulate the accumulation of the ramp
    ramp_value = np.zeros(n_pixels, dtype=np.float32)  # Last value of ramp
    for index, reads in enumerate(read_pattern):
        resultant_total = np.zeros(n_pixels, dtype=np.float32)  # Total of all reads in this resultant
        for _ in reads:
            # Compute the next value of the ramp
            #   Using a Poisson process for the flux
            ramp_value += RNG.poisson(FLUX * READ_TIME, size=n_pixels).astype(np.float32)

            # Add to running total for the resultant
            resultant_total += ramp_value

        # Add read noise to the resultant
        resultant_total += (
            RNG.standard_normal(size=n_pixels, dtype=np.float32) * READ_NOISE / np.sqrt(len(reads))
        )

        # Record the average value for resultant (i.e., the average of the reads)
        resultants[index] = (resultant_total / len(reads)).astype(np.float32)

    if n_pixels == 1:
        resultants = resultants[:, 0]

    return resultants


@pytest.fixture(scope="module")
def pixel_data(ramp_data):
    """
    Create data for a single pixel

    Returns:
        resultants
            Resultants for a single pixel
        t_bar:
            The t_bar values for the read pattern used for the resultants
        tau:
            The tau values for the read pattern used for the resultants
        n_reads:
            The number of reads for the read pattern used for the resultants
    """
    t_bar, tau, n_reads, read_pattern = ramp_data

    n_resultants = len(t_bar)
    fixed = np.empty((FixedOffsets.n_fixed_offsets, n_resultants - 1), dtype=np.float32)
    fixed = fill_fixed_values(fixed, t_bar, tau, n_reads, n_resultants)

    resultants = _generate_resultants(read_pattern)

    return resultants, t_bar, tau, n_reads, fixed


def test__fill_pixel_values(pixel_data):
    """Test computing the initial pixel data"""
    resultants, t_bar, tau, n_reads, fixed = pixel_data

    n_resultants = len(t_bar)
    pixel = np.empty((PixelOffsets.n_pixel_offsets, n_resultants - 1), dtype=np.float32)
    pixel = _fill_pixel_values(pixel, resultants, fixed, READ_NOISE, n_resultants)

    # Sanity check that the shape of pixel is correct
    assert pixel.shape == (2 * 2, n_resultants - 1)

    # Split into the different types of data
    local_slopes = pixel[PixelOffsets.single_local_slope : PixelOffsets.double_local_slope + 1, :]
    var_read_noise = pixel[PixelOffsets.single_var_read_noise : PixelOffsets.double_var_read_noise + 1, :]

    # Sanity check that these are all the right shape
    assert local_slopes.shape == (2, n_resultants - 1)
    assert var_read_noise.shape == (2, n_resultants - 1)

    # Check the computed data
    #   These are computed using loop in cython, here we check against numpy
    # Single diffs
    assert np.all(local_slopes[0] == (resultants[1:] - resultants[:-1]) / (t_bar[1:] - t_bar[:-1]))
    assert np.all(
        var_read_noise[0]
        == np.float32(READ_NOISE**2) * (np.float32(1 / n_reads[1:]) + np.float32(1 / n_reads[:-1]))
    )

    # Double diffs
    assert np.all(local_slopes[1, :-1] == (resultants[2:] - resultants[:-2]) / (t_bar[2:] - t_bar[:-2]))
    assert np.all(
        var_read_noise[1, :-1]
        == np.float32(READ_NOISE**2) * (np.float32(1 / n_reads[2:]) + np.float32(1 / n_reads[:-2]))
    )

    # Last double diff should be NaN
    assert np.isnan(local_slopes[1, -1])
    assert np.isnan(var_read_noise[1, -1])


@pytest.fixture(scope="module")
def detector_data(read_pattern):
    """
    Generate a set of with no jumps data as if for a single detector as it
        would be passed in by the supporting code.

    Returns:
        resultants
            The resultants for a large number of pixels
        read_noise:
            The read noise vector for those pixels
        read_pattern:
            The read pattern used for the resultants
    """
    read_noise = np.ones(N_PIXELS, dtype=np.float32) * READ_NOISE

    resultants = _generate_resultants(read_pattern, n_pixels=N_PIXELS)

    return resultants, read_noise, read_pattern


@pytest.mark.parametrize("use_jump", [True, False])
@pytest.mark.parametrize("use_dq", [True, False])
def test_fit_ramps(detector_data, use_jump, use_dq):
    """
    Test fitting ramps
        Since no jumps are simulated in the data, jump detection shouldn't pick
        up any jumps.
    """
    resultants, read_noise, read_pattern = detector_data
    dq = (
        (RNG.uniform(size=resultants.shape) > GOOD_PROB).astype(np.int32)
        if use_dq
        else np.zeros(resultants.shape, dtype=np.int32)
    )

    # only use okay ramps
    #   ramps passing the below criterion have at least two adjacent valid reads
    #   i.e., we can make a measurement from them.
    okay = np.sum((dq[1:, :] == 0) & (dq[:-1, :] == 0), axis=0) != 0
    assert okay.dtype == bool

    # Note that for use_dq = False, okay == True for all ramps, so we perform
    #    a sanity check that the above criterion is correct
    if not use_dq:
        assert okay.all()

    output = fit_ramps(
        resultants, dq, read_noise, READ_TIME, read_pattern, use_jump=use_jump, include_diagnostic=True
    )
    assert len(output.fits) == N_PIXELS  # sanity check that a fit is output for each pixel

    chi2 = 0
    for fit, use in zip(output.fits, okay):
        if not use_dq and not use_jump:
            ##### The not use_jump makes this NOT test for false positives #####
            # Check that the data generated does not generate any false positives
            #   for jumps as this data is reused for `test_find_jumps` below.
            #   This guarantees that all jumps detected in that test are the
            #   purposefully placed ones which we know about. So the `test_find_jumps`
            #   can focus on checking that the jumps found are the correct ones,
            #   and that all jumps introduced are detected properly.
            assert len(fit["fits"]) == 1

        if use:
            # Add okay ramps to chi2
            total_var = fit["average"]["read_var"] + fit["average"]["poisson_var"]
            if total_var != 0:
                chi2 += (fit["average"]["slope"] - FLUX) ** 2 / total_var
        else:
            # Check no slope fit for bad ramps
            assert fit["average"]["slope"] == 0
            assert fit["average"]["read_var"] == 0
            assert fit["average"]["poisson_var"] == 0

            assert use_dq  # sanity check that this branch is only encountered when use_dq = True

    chi2 /= np.sum(okay)
    assert np.abs(chi2 - 1) < CHI2_TOL


@pytest.mark.parametrize("use_jump", [True, False])
def test_fit_ramps_array_outputs(detector_data, use_jump):
    """
    Test that the array outputs line up with the dictionary output
    """
    resultants, read_noise, read_pattern = detector_data
    dq = np.zeros(resultants.shape, dtype=np.int32)

    output = fit_ramps(
        resultants, dq, read_noise, READ_TIME, read_pattern, use_jump=use_jump, include_diagnostic=True
    )

    for fit, par, var in zip(output.fits, output.parameters, output.variances):
        assert par[Parameter.intercept] == 0
        assert par[Parameter.slope] == fit["average"]["slope"]

        assert var[Variance.read_var] == fit["average"]["read_var"]
        assert var[Variance.poisson_var] == fit["average"]["poisson_var"]
        assert var[Variance.total_var] == np.float32(
            fit["average"]["read_var"] + fit["average"]["poisson_var"]
        )


@pytest.fixture(scope="module")
def jump_data(detector_data):
    """
    Generate resultants with single jumps in them for testing jump detection.
        Note this specifically checks that we can detect jumps in any read, meaning
        it has an insurance check that a jump has been placed in every single
        read position.

    Returns:
        resultants
            The resultants for a large number of pixels
        read_noise:
            The read noise vector for those pixels
        read_pattern:
            The read pattern used for the resultants
        jump_reads:
            Index of read where a jump occurs for each pixel
        jump_resultants:
            Index of resultant where a jump occurs for each pixel
    """
    resultants, read_noise, read_pattern = detector_data

    # Choose read to place a single jump in for each pixel
    num_reads = read_pattern[-1][-1]
    jump_reads = RNG.integers(num_reads - 1, size=N_PIXELS)

    # This shows that a jump as been placed in every single possible
    #   read position. Technically, this check can fail; however,
    #   N_PIXELS >> num_reads so it is very unlikely in practice since
    #   all reads are equally likely to be chosen for a jump.
    # It is a good check that we can detect a jump occurring in any read except
    #    the first read.
    assert set(jump_reads) == set(range(num_reads - 1))

    # Fill out jump reads with jump values
    jump_flux = np.zeros((num_reads, N_PIXELS), dtype=np.float32)
    for index, jump in enumerate(jump_reads):
        jump_flux[jump:, index] = JUMP_VALUE

    # Average the reads into the resultants
    jump_resultants = np.zeros(N_PIXELS, dtype=np.int32)
    for index, reads in enumerate(read_pattern):
        indices = np.array(reads) - 1
        resultants[index, :] += np.mean(jump_flux[indices, :], axis=0)
        for read in reads:
            jump_resultants[np.where(jump_reads == read)] = index

    return resultants, read_noise, read_pattern, jump_reads, jump_resultants


def test_find_jumps(jump_data):
    """
    Full unit tests to demonstrate that we can detect jumps in any read (except
    the first one) and that we correctly remove these reads from the fit to recover
    the correct FLUX/slope.
    """
    resultants, read_noise, read_pattern, jump_reads, jump_resultants = jump_data
    dq = np.zeros(resultants.shape, dtype=np.int32)

    output = fit_ramps(
        resultants, dq, read_noise, READ_TIME, read_pattern, use_jump=True, include_diagnostic=True
    )
    assert len(output.fits) == len(jump_reads)  # sanity check that a fit/jump is set for every pixel

    chi2 = 0
    incorrect_too_few = 0
    incorrect_too_many = 0
    incorrect_does_not_capture = 0
    incorrect_other = 0
    for fit, jump_index, resultant_index in zip(output.fits, jump_reads, jump_resultants):
        # Check that the jumps are detected correctly
        if jump_index == 0:
            # There is no way to detect a jump if it is in the very first read
            # The very first pixel in this case has a jump in the first read
            assert len(fit["jumps"]) == 0
            assert resultant_index == 0  # sanity check that the jump is indeed in the first resultant

            # Test the correct ramp_index was recorded:
            assert len(fit["index"]) == 1
            assert fit["index"][0]["start"] == 0
            assert fit["index"][0]["end"] == len(read_pattern) - 1
        else:
            # There should be a single jump detected; however, this results in
            # two resultants being excluded.
            if resultant_index not in fit["jumps"]:
                incorrect_does_not_capture += 1
                continue
            if len(fit["jumps"]) < 2:
                incorrect_too_few += 1
                continue
            if len(fit["jumps"]) > 2:
                incorrect_too_many += 1
                continue

            # The two resultants excluded should be adjacent
            jump_correct = [
                (jump in (resultant_index, resultant_index - 1, resultant_index + 1)) for jump in fit["jumps"]
            ]
            if not all(jump_correct):
                incorrect_other += 1
                continue

            # Because we do not have a data set with no false positives, we cannot run the below
            # # Test the correct ramp indexes are recorded
            # ramp_indices = []
            # for ramp_index in fit['index']:
            #     # Note start/end of a ramp_index are inclusive meaning that end
            #     #    is an index included in the ramp_index so the range is to end + 1
            #     new_indices = list(range(ramp_index["start"], ramp_index["end"] + 1))

            #     # check that all the ramps are non-overlapping
            #     assert set(ramp_indices).isdisjoint(new_indices)

            #     ramp_indices.extend(new_indices)

            # # check that no ramp_index is a jump
            # assert set(ramp_indices).isdisjoint(fit['jumps'])

            # # check that all resultant indices are either in a ramp or listed as a jump
            # assert set(ramp_indices).union(fit['jumps']) == set(range(len(read_pattern)))

        # Compute the chi2 for the fit and add it to a running "total chi2"
        total_var = fit["average"]["read_var"] + fit["average"]["poisson_var"]
        chi2 += (fit["average"]["slope"] - FLUX) ** 2 / total_var

    # Check that the average chi2 is ~1.
    chi2 /= N_PIXELS - incorrect_too_few - incorrect_too_many - incorrect_does_not_capture - incorrect_other
    assert np.abs(chi2 - 1) < CHI2_TOL


def test_override_default_threshold(jump_data):
    """This tests that we can override the default jump detection threshold constants"""
    resultants, read_noise, read_pattern, jump_reads, jump_resultants = jump_data
    dq = np.zeros(resultants.shape, dtype=np.int32)

    standard = fit_ramps(resultants, dq, read_noise, READ_TIME, read_pattern, use_jump=True)
    override = fit_ramps(
        resultants, dq, read_noise, READ_TIME, read_pattern, use_jump=True, intercept=0, constant=0
    )

    # All this is intended to do is show that with all other things being equal passing non-default
    #    threshold parameters changes the results.
    assert (standard.parameters != override.parameters).any()


def test_jump_dq_set(jump_data):
    # Check the DQ flag value to start
    assert 2**2 == JUMP_DET

    resultants, read_noise, read_pattern, jump_reads, jump_resultants = jump_data
    dq = np.zeros(resultants.shape, dtype=np.int32)

    output = fit_ramps(
        resultants, dq, read_noise, READ_TIME, read_pattern, use_jump=True, include_diagnostic=True
    )

    for fit, pixel_dq in zip(output.fits, output.dq.transpose()):
        # Check that all jumps found get marked
        assert (pixel_dq[fit["jumps"]] == JUMP_DET).all()

        # Check that dq flags for jumps are only set if the jump is marked
        assert set(np.where(pixel_dq == JUMP_DET)[0]) == set(fit["jumps"])



import logging
import warnings

import numpy as np
import warnings
from astropy import stats

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def find_crs(
    dataa,
    group_dq,
    read_noise,
    normal_rej_thresh,
    two_diff_rej_thresh,
    three_diff_rej_thresh,
    nframes,
    flag_4_neighbors,
    max_jump_to_flag_neighbors,
    min_jump_to_flag_neighbors,
    dqflags,
    after_jump_flag_e1=0.0,
    after_jump_flag_n1=0,
    after_jump_flag_e2=0.0,
    after_jump_flag_n2=0,
    copy_arrs=True,
    minimum_groups=3,
    minimum_sigclip_groups=100,
    only_use_ints=True,
    min_diffs_single_pass=10,
):
    """
    Find CRs/Jumps in each integration within the input data array. The input
    data array is assumed to be in units of electrons, i.e. already multiplied
    by the gain. We also assume that the read noise is in units of electrons.
    We also assume that there are at least three groups in the integrations.
    This was checked by jump_step before this routine is called.

    Parameters
    ----------
    dataa: float, 4D array (num_ints, num_groups, num_rows,  num_cols)
        input ramp data

    group_dq : int, 4D array
        group DQ flags

    read_noise : float, 2D array
        The read noise of each pixel

    normal_rej_thresh : float
        cosmic ray sigma rejection threshold

    two_diff_rej_thresh : float
        cosmic ray sigma rejection threshold for ramps having 3 groups

    three_diff_rej_thresh : float
        cosmic ray sigma rejection threshold for ramps having 4 groups

    nframes : int
        The number of frames that are included in the group average

    flag_4_neighbors : bool
        if set to True (default is True), it will cause the four perpendicular
        neighbors of all detected jumps to also be flagged as a jump.

    max_jump_to_flag_neighbors : float
        value in units of sigma that sets the upper limit for flagging of
        neighbors. Any jump above this cutoff will not have its neighbors
        flagged.

    min_jump_to_flag_neighbors : float
        value in units of sigma that sets the lower limit for flagging of
        neighbors (marginal detections). Any primary jump below this value will
        not have its neighbors flagged.

    dqflags: dict
        A dictionary with at least the following keywords:
        DO_NOT_USE, SATURATED, JUMP_DET, NO_GAIN_VALUE, GOOD

    after_jump_flag_e1 : float
        Jumps with amplitudes above the specified e value will have subsequent
        groups flagged with the number determined by the after_jump_flag_n1

    after_jump_flag_n1 : int
        Gives the number of groups to flag after jumps with DN values above that
        given by after_jump_flag_dn1

    after_jump_flag_e2 : float
        Jumps with amplitudes above the specified e value will have subsequent
        groups flagged with the number determined by the after_jump_flag_n2

    after_jump_flag_n2 : int
        Gives the number of groups to flag after jumps with DN values above that
        given by after_jump_flag_dn2

    copy_arrs : bool
        Flag for making internal copies of the arrays so the input isn't modified,
        defaults to True.

    minimum_groups : integer
        The minimum number of groups to perform jump detection.

    minimum_sigclip_groups : integer
        The minimum number of groups required for the sigma clip routine to be
        used for jump detection rather than using the expected noise based on
        the read noise and gain files.

    only_use_ints : boolean
        If True the sigma clip process will only apply for groups between
        integrations. This means that a group will only be compared against the
        same group in other integrations. If False all groups across all integrations
        will be used to detect outliers.

    min_diffs_single_pass: integer
        The minimum number of groups to switch from the iterative flagging of
        cosmic rays to just finding all the outliers at once.
    Returns
    -------
    gdq : int, 4D array
        group DQ array with reset flags

    row_below_gdq : int, 3D array (num_ints, num_groups, num_cols)
        pixels below current row also to be flagged as a CR

    row_above_gdq : int, 3D array (num_ints, num_groups, num_cols)
        pixels above current row also to be flagged as a CR

    """
    # copy data and group DQ array
    if copy_arrs:
        dat = dataa.copy()
        gdq = group_dq.copy()
    else:
        dat = dataa
        gdq = group_dq
    # Get data characteristics
    nints, ngroups, nrows, ncols = dataa.shape
    ndiffs = (ngroups - 1) * nints
    # get readnoise, squared
    read_noise_2 = read_noise**2
    # create arrays for output
    row_above_gdq = np.zeros((nints, ngroups, ncols), dtype=np.uint8)
    row_below_gdq = np.zeros((nints, ngroups, ncols), dtype=np.uint8)

    # get dq flags for saturated, donotuse, jump
    sat_flag = dqflags["SATURATED"]
    dnu_flag = dqflags["DO_NOT_USE"]
    jump_flag = dqflags["JUMP_DET"]

    # get data, gdq
    num_flagged_grps = 0
    # determine the number of groups with all pixels set to DO_NOT_USE
    ngrps = dat.shape[1]
    for integ in range(nints):
        for grp in range(dat.shape[1]):
            if np.all(np.bitwise_and(gdq[integ, grp, :, :], dnu_flag)):
                num_flagged_grps += 1
    if only_use_ints:
        total_sigclip_groups = nints
    else:
        total_sigclip_groups = nints * (ngrps - num_flagged_grps)
    total_groups = nints * (ngrps - num_flagged_grps)
    total_diffs = nints * (ngrps - 1 - num_flagged_grps)
    total_usable_diffs = total_diffs - num_flagged_grps
    if ((ngrps < minimum_groups and only_use_ints and nints < minimum_sigclip_groups) or
            (not only_use_ints and nints * ngrps < minimum_sigclip_groups and
             total_groups < minimum_groups)):
        log.info("Jump Step was skipped because exposure has less than the minimum number of usable groups")
        dummy = np.zeros((dataa.shape[1] - 1, dataa.shape[2], dataa.shape[3]),
                         dtype=np.float32)
        return gdq, row_below_gdq, row_above_gdq, 0, dummy
    else:
        # set 'saturated' or 'do not use' pixels to nan in data
        dat[np.where(np.bitwise_and(gdq, sat_flag))] = np.nan
        dat[np.where(np.bitwise_and(gdq, dnu_flag))] = np.nan
        dat[np.where(np.bitwise_and(gdq, dnu_flag + sat_flag))] = np.nan

        # calculate the differences between adjacent groups (first diffs)
        # use mask on data, so the results will have sat/donotuse groups masked
        first_diffs = np.diff(dat, axis=1)

        # calc. the median of first_diffs for each pixel along the group axis
        first_diffs_masked = np.ma.masked_array(first_diffs, mask=np.isnan(first_diffs))
        median_diffs = np.ma.median(first_diffs_masked, axis=(0, 1))
        # calculate sigma for each pixel
        sigma = np.sqrt(np.abs(median_diffs) + read_noise_2 / nframes)

        # reset sigma so pxels with 0 readnoise are not flagged as jumps
        sigma[np.where(sigma == 0.)] = np.nan

        # compute 'ratio' for each group. this is the value that will be
        # compared to 'threshold' to classify jumps. subtract the median of
        # first_diffs from first_diffs, take the absolute value and divide by sigma.
        e_jump_4d = first_diffs - median_diffs[np.newaxis, :, :]
        ratio_all = np.abs(first_diffs - median_diffs[np.newaxis, np.newaxis, :, :]) / \
                    sigma[np.newaxis, np.newaxis, :, :]
        # Test to see if there are enough groups to use sigma clipping
        if (only_use_ints and nints >= minimum_sigclip_groups) or \
           (not only_use_ints and total_groups >= minimum_sigclip_groups):
            log.info(" Jump Step using sigma clip {} greater than {}, rejection threshold {}".format(
                str(total_groups), str(minimum_sigclip_groups), str(normal_rej_thresh)))
            warnings.filterwarnings("ignore", ".*All-NaN slice encountered.*", RuntimeWarning)
            warnings.filterwarnings("ignore", ".*Mean of empty slice.*", RuntimeWarning)
            warnings.filterwarnings("ignore", ".*Degrees of freedom <= 0.*", RuntimeWarning)

            if only_use_ints:
                mean, median, stddev = stats.sigma_clipped_stats(first_diffs_masked, sigma=normal_rej_thresh,
                                                                 axis=0)
                clipped_diffs = stats.sigma_clip(first_diffs_masked, sigma=normal_rej_thresh,
                                                 axis=0, masked=True)
            else:
                mean, median, stddev = stats.sigma_clipped_stats(first_diffs_masked, sigma=normal_rej_thresh,
                                                                 axis=(0, 1))
                clipped_diffs = stats.sigma_clip(first_diffs_masked, sigma=normal_rej_thresh,
                                                 axis=(0, 1), masked=True)
            jump_mask = np.logical_and(clipped_diffs.mask, np.logical_not(first_diffs_masked.mask))
            jump_mask[np.bitwise_and(jump_mask, gdq[:, 1:, :, :] == sat_flag)] = False
            jump_mask[np.bitwise_and(jump_mask, gdq[:, 1:, :, :] == dnu_flag)] = False
            jump_mask[np.bitwise_and(jump_mask, gdq[:, 1:, :, :] == (dnu_flag + sat_flag))] = False
            gdq[:, 1:, :, :] = np.bitwise_or(gdq[:, 1:, :, :], jump_mask *
                                             np.uint8(dqflags["JUMP_DET"]))
            # if grp is all jump set to do not use
            for integ in range(nints):
                for grp in range(ngrps):
                    if np.all(np.bitwise_or(np.bitwise_and(gdq[integ, grp, :, :], jump_flag),
                                            np.bitwise_and(gdq[integ, grp, :, :], dnu_flag))):
                        jumpy, jumpx = np.where(gdq[integ, grp, :, :] == jump_flag)
                        gdq[integ, grp, jumpy, jumpx] = 0
            warnings.resetwarnings()
        else:  # There are not enough groups for sigma clipping

                # set 'saturated' or 'do not use' pixels to nan in data
                dat[np.where(np.bitwise_and(gdq, sat_flag))] = np.nan
                dat[np.where(np.bitwise_and(gdq, dnu_flag))] = np.nan

                # calculate the differences between adjacent groups (first diffs)
                # use mask on data, so the results will have sat/donotuse groups masked
                first_diffs = np.diff(dat, axis=1)

                if total_usable_diffs >= min_diffs_single_pass:
                    warnings.filterwarnings("ignore", ".*All-NaN slice encountered.*", RuntimeWarning)
                    median_diffs = np.nanmedian(first_diffs, axis=(0, 1))
                    warnings.resetwarnings()
                    # calculate sigma for each pixel
                    sigma = np.sqrt(np.abs(median_diffs) + read_noise_2 / nframes)
                    # reset sigma so pixels with 0 read noise are not flagged as jumps
                    sigma[np.where(sigma == 0.)] = np.nan

                    # compute 'ratio' for each group. this is the value that will be
                    # compared to 'threshold' to classify jumps. subtract the median of
                    # first_diffs from first_diffs, take the abs. value and divide by sigma.
                    e_jump = first_diffs - median_diffs[np.newaxis, np.newaxis, :, :]

                    ratio = np.abs(e_jump) / sigma[np.newaxis, np.newaxis, :, :]
                    masked_ratio = np.ma.masked_greater(ratio, normal_rej_thresh)
                    #  The jump mask is the ratio greater than the threshold and the difference is usable
                    jump_mask = np.logical_and(masked_ratio.mask, np.logical_not(first_diffs_masked.mask))
                    gdq[:, 1:, :, :] = np.bitwise_or(gdq[:, 1:, :, :], jump_mask *
                                                     np.uint8(dqflags["JUMP_DET"]))
                else:  # low number of diffs requires iterative flagging
                    # calculate the differences between adjacent groups (first diffs)
                    # use mask on data, so the results will have sat/donotuse groups masked
                    first_diffs = np.abs(np.diff(dat, axis=1))

                    # calc. the median of first_diffs for each pixel along the group axis
                    median_diffs = calc_med_first_diffs(first_diffs)

                    # calculate sigma for each pixel
                    sigma = np.sqrt(np.abs(median_diffs) + read_noise_2 / nframes)
                    # reset sigma so pxels with 0 readnoise are not flagged as jumps
                    sigma[np.where(sigma == 0.0)] = np.nan

                    # compute 'ratio' for each group. this is the value that will be
                    # compared to 'threshold' to classify jumps. subtract the median of
                    # first_diffs from first_diffs, take the abs. value and divide by sigma.
                    e_jump = first_diffs - median_diffs[np.newaxis, :, :]
                    ratio = np.abs(e_jump) / sigma[np.newaxis, :, :]

                    # create a 2d array containing the value of the largest 'ratio' for each pixel
                    warnings.filterwarnings("ignore", ".*All-NaN slice encountered.*", RuntimeWarning)
                    max_ratio = np.nanmax(ratio, axis=1)
                    warnings.resetwarnings()
                    # now see if the largest ratio of all groups for each pixel exceeds the threshold.
                    # there are different threshold for 4+, 3, and 2 usable groups
                    num_unusable_groups = np.sum(np.isnan(first_diffs), axis=(0, 1))
                    int4cr, row4cr, col4cr = np.where(
                        np.logical_and(ndiffs - num_unusable_groups >= 4, max_ratio > normal_rej_thresh)
                    )
                    int3cr, row3cr, col3cr = np.where(
                        np.logical_and(ndiffs - num_unusable_groups == 3, max_ratio > three_diff_rej_thresh)
                    )
                    int2cr, row2cr, col2cr = np.where(
                        np.logical_and(ndiffs - num_unusable_groups == 2, max_ratio > two_diff_rej_thresh)
                    )
                    # get the rows, col pairs for all pixels with at least one CR
#                    all_crs_int = np.concatenate((int4cr, int3cr, int2cr))
                    all_crs_row = np.concatenate((row4cr, row3cr, row2cr))
                    all_crs_col = np.concatenate((col4cr, col3cr, col2cr))

                    # iterate over all groups of the pix w/ an initial CR to look for subsequent CRs
                    # flag and clip the first CR found. recompute median/sigma/ratio
                    # and repeat the above steps of comparing the max 'ratio' for each pixel
                    # to the threshold to determine if another CR can be flagged and clipped.
                    # repeat this process until no more CRs are found.
                    for j in range(len(all_crs_row)):
                        # get arrays of abs(diffs), ratio, readnoise for this pixel
                        pix_first_diffs = first_diffs[:, :, all_crs_row[j], all_crs_col[j]]
                        pix_ratio = ratio[:, :, all_crs_row[j], all_crs_col[j]]
                        pix_rn2 = read_noise_2[all_crs_row[j], all_crs_col[j]]
    
                        # Create a mask to flag CRs. pix_cr_mask = 0 denotes a CR
                        pix_cr_mask = np.ones(pix_first_diffs.shape, dtype=bool)

                        # set the largest ratio as a CR
                        location = np.unravel_index(np.nanargmax(pix_ratio), pix_ratio.shape)
                        pix_cr_mask[location] = 0
                        new_CR_found = True

                        # loop and check for more CRs, setting the mask as you go and
                        # clipping the group with the CR. stop when no more CRs are found
                        # or there is only one two diffs left (which means there is
                        # actually one left, since the next CR will be masked after
                        # checking that condition)
                        while new_CR_found and (ndiffs - np.sum(np.isnan(pix_first_diffs)) > 2):
                            new_CR_found = False

                            # set CRs to nans in first diffs to clip them
                            pix_first_diffs[~pix_cr_mask] = np.nan

                            # recalculate median, sigma, and ratio
                            new_pix_median_diffs = calc_med_first_diffs(pix_first_diffs)

                            new_pix_sigma = np.sqrt(np.abs(new_pix_median_diffs) + pix_rn2 / nframes)
                            new_pix_ratio = np.abs(pix_first_diffs - new_pix_median_diffs) / new_pix_sigma

                            # check if largest ratio exceeds threshold appropriate for num remaining groups

                            # select appropriate thresh. based on number of remaining groups
                            rej_thresh = normal_rej_thresh
                            if ndiffs - np.sum(np.isnan(pix_first_diffs)) == 3:
                                rej_thresh = three_diff_rej_thresh
                            if ndiffs - np.sum(np.isnan(pix_first_diffs)) == 2:
                                rej_thresh = two_diff_rej_thresh
                            max_idx = np.nanargmax(new_pix_ratio)
                            location = np.unravel_index(max_idx, new_pix_ratio.shape)
                            if new_pix_ratio[location] > rej_thresh:
                                new_CR_found = True
                                pix_cr_mask[location] = 0
                            unusable_diffs = np.sum(np.isnan(pix_first_diffs))
                        # Found all CRs for this pix - set flags in input DQ array
                        gdq[:, 1:, all_crs_row[j], all_crs_col[j]] = np.bitwise_or(
                             gdq[:, 1:, all_crs_row[j], all_crs_col[j]],
                            dqflags["JUMP_DET"] * np.invert(pix_cr_mask),
                        )
    cr_integ, cr_group, cr_row, cr_col = np.where(np.bitwise_and(gdq, jump_flag))
    num_primary_crs = len(cr_group)
    if flag_4_neighbors:  # iterate over each 'jump' pixel
        for j in range(len(cr_group)):
            ratio_this_pix = ratio_all[cr_integ[j], cr_group[j] - 1, cr_row[j], cr_col[j]]

            # Jumps must be in a certain range to have neighbors flagged
            if (ratio_this_pix < max_jump_to_flag_neighbors) and (
                ratio_this_pix > min_jump_to_flag_neighbors
            ):
                integ = cr_integ[j]
                group = cr_group[j]
                row = cr_row[j]
                col = cr_col[j]

                # This section saves flagged neighbors that are above or
                # below the current range of row. If this method
                # running in a single process, the row above and below are
                # not used. If it is running in multiprocessing mode, then
                # the rows above and below need to be returned to
                # find_jumps to use when it reconstructs the full group dq
                # array from the slices.

                # Only flag adjacent pixels if they do not already have the
                # 'SATURATION' or 'DONOTUSE' flag set
                if row != 0:
                    if (gdq[integ, group, row - 1, col] & sat_flag) == 0 and (
                        gdq[integ, group, row - 1, col] & dnu_flag
                    ) == 0:
                        gdq[integ, group, row - 1, col] = np.bitwise_or(
                            gdq[integ, group, row - 1, col], jump_flag
                        )
                else:
                    row_below_gdq[integ, cr_group[j], cr_col[j]] = jump_flag

                if row != nrows - 1:
                    if (gdq[integ, group, row + 1, col] & sat_flag) == 0 and (
                        gdq[integ, group, row + 1, col] & dnu_flag
                    ) == 0:
                        gdq[integ, group, row + 1, col] = np.bitwise_or(
                            gdq[integ, group, row + 1, col], jump_flag
                        )
                else:
                    row_above_gdq[integ, cr_group[j], cr_col[j]] = jump_flag

                # Here we are just checking that we don't flag neighbors of
                # jumps that are off the detector.
                if (
                    cr_col[j] != 0
                    and (gdq[integ, group, row, col - 1] & sat_flag) == 0
                    and (gdq[integ, group, row, col - 1] & dnu_flag) == 0
                ):
                    gdq[integ, group, row, col - 1] = np.bitwise_or(
                        gdq[integ, group, row, col - 1], jump_flag
                    )

                if (
                    cr_col[j] != ncols - 1
                    and (gdq[integ, group, row, col + 1] & sat_flag) == 0
                    and (gdq[integ, group, row, col + 1] & dnu_flag) == 0
                ):
                    gdq[integ, group, row, col + 1] = np.bitwise_or(
                        gdq[integ, group, row, col + 1], jump_flag
                    )

    # flag n groups after jumps above the specified thresholds to account for
    # the transient seen after ramp jumps
    flag_e_threshold = [after_jump_flag_e1, after_jump_flag_e2]
    flag_groups = [after_jump_flag_n1, after_jump_flag_n2]
    for cthres, cgroup in zip(flag_e_threshold, flag_groups):
        if cgroup > 0:
            cr_intg, cr_group, cr_row, cr_col = np.where(np.bitwise_and(gdq, jump_flag))
            for j in range(len(cr_group)):
                intg = cr_intg[j]
                group = cr_group[j]
                row = cr_row[j]
                col = cr_col[j]
                if e_jump_4d[intg, group - 1, row, col] >= cthres:
                    for kk in range(group, min(group + cgroup + 1, ngroups)):
                        if (gdq[intg, kk, row, col] & sat_flag) == 0 and (
                            gdq[intg, kk, row, col] & dnu_flag
                        ) == 0:
                            gdq[intg, kk, row, col] = np.bitwise_or(gdq[intg, kk, row, col], jump_flag)

    if "stddev" in locals():
        return gdq, row_below_gdq, row_above_gdq, num_primary_crs, stddev

    if only_use_ints:
        dummy = np.zeros((dataa.shape[1] - 1, dataa.shape[2], dataa.shape[3]), dtype=np.float32)
    else:
        dummy = np.zeros((dataa.shape[2], dataa.shape[3]), dtype=np.float32)

    return gdq, row_below_gdq, row_above_gdq, num_primary_crs, dummy


def calc_med_first_diffs(in_first_diffs):
    """Calculate the median of `first diffs` along the group axis.

    If there are 4+ usable groups (e.g not flagged as saturated, donotuse,
    or a previously clipped CR), then the group with largest absolute
    first difference will be clipped and the median of the remaining groups
    will be returned. If there are exactly 3 usable groups, the median of
    those three groups will be returned without any clipping. Finally, if
    there are two usable groups, the group with the smallest absolute
    difference will be returned.
    Parameters
    ----------
    in_first_diffs : array, float
        array containing the first differences of adjacent groups
        for a single integration. Can be 3d or 1d (for a single pix)

    Returns
    -------
    median_diffs : float or array, float
        If the input is a single pixel, a float containing the median for
        the groups in that pixel will be returned. If the input is a 3d
        array of several pixels, a 2d array with the median for each pixel
        will be returned.
    """
    first_diffs = in_first_diffs.copy()
    if first_diffs.ndim == 1:  # in the case where input is a single pixel
        num_usable_groups = len(first_diffs) - np.sum(np.isnan(first_diffs), axis=0)
        if num_usable_groups >= 4:  # if 4+, clip largest and return median
            mask = np.ones_like(first_diffs).astype(bool)
            mask[np.nanargmax(np.abs(first_diffs))] = False  # clip the diff with the largest abs value
            return np.nanmedian(first_diffs[mask])

        if num_usable_groups == 3:  # if 3, no clipping just return median
            return np.nanmedian(first_diffs)

        if num_usable_groups == 2:  # if 2, return diff with minimum abs
            return first_diffs[np.nanargmin(np.abs(first_diffs))]

        return np.nan

    if first_diffs.ndim == 2:  # in the case where input is a single pixel
        nansum = np.sum(np.isnan(first_diffs), axis=(0, 1))
        num_usable_diffs = first_diffs.size - np.sum(np.isnan(first_diffs), axis=(0, 1))
        if num_usable_diffs >= 4:  # if 4+, clip largest and return median
            mask = np.ones_like(first_diffs).astype(bool)
            location = np.unravel_index(first_diffs.argmax(), first_diffs.shape)
            mask[location] = False  # clip the diff with the largest abs value
            return np.nanmedian(first_diffs[mask])
        elif num_usable_diffs == 3:  # if 3, no clipping just return median
            return np.nanmedian(first_diffs)
        elif num_usable_diffs == 2:  # if 2, return diff with minimum abs
            TEST = np.nanargmin(np.abs(first_diffs))
            diff_min_idx = np.nanargmin(first_diffs)
            location = np.unravel_index(diff_min_idx, first_diffs.shape)
            return first_diffs[location]
        else:
            return np.nan

    if first_diffs.ndim == 4:
        # if input is multi-dimensional
        nints, ndiffs, nrows, ncols = first_diffs.shape
        shaped_diffs = np.reshape(first_diffs, ((nints * ndiffs), nrows, ncols))
        num_usable_diffs = (ndiffs * nints) - np.sum(np.isnan(shaped_diffs), axis=0)
        median_diffs = np.zeros((nrows, ncols))  # empty array to store median for each pix

        # process groups with >=4 usable diffs
        row4, col4 = np.where(num_usable_diffs >= 4)  # locations of >= 4 usable diffs pixels
        if len(row4) > 0:
            four_slice = shaped_diffs[:, row4, col4]
            loc0 = np.nanargmax(four_slice, axis=0)
            shaped_diffs[loc0, row4, col4] = np.nan
            median_diffs[row4, col4] = np.nanmedian(shaped_diffs[:, row4, col4], axis=0)

        # process groups with 3 usable groups
        row3, col3 = np.where(num_usable_diffs == 3)  # locations of == 3 usable diff pixels
        if len(row3) > 0:
            three_slice = shaped_diffs[:, row3, col3]
            median_diffs[row3, col3] = np.nanmedian(three_slice, axis=0)  # add median to return arr for these pix

        # process groups with 2 usable groups
        row2, col2 = np.where(num_usable_diffs == 2)  # locations of == 2 usable diff pixels
        if len(row2) > 0:
            two_slice = shaped_diffs[ :, row2, col2]
            two_slice[np.nanargmax(np.abs(two_slice), axis=0),
                      np.arange(two_slice.shape[1])] = np.nan  # mask larger abs. val
            median_diffs[row2, col2] = np.nanmin(two_slice, axis=0)  # add med. to return arr

        # set the medians all groups with less than 2 usable diffs to nan to skip further
        # calculations for these pixels
        row_none, col_none = np.where(num_usable_diffs < 2)
        median_diffs[row_none, col_none] = np.nan

        return median_diffs



import inspect
from pathlib import Path

import pytest
import numpy as np
import numpy.testing as npt

from stcal.ramp_fitting.ramp_fit import ramp_fit_data
from stcal.ramp_fitting.ramp_fit_class import RampData

#
# The first 12 tests are for a single ramp in a single integration. The ramps
#  have a variety of GROUPDQ vectors, with 1 or more segments in each ramp.  The
#  comparison of the PRIMARY output results are partly to verify the slopes and
#  variances of the combination of the segments in a ramp within the single
#  integration.  The comparison of the OPTIONAL output results are to verify the
#  results for each of the individual segments in a ramp.  Within each test is a
#  description of classification ('CASE') within the code of all of the segments
#  for the pixel based on the ramp's GROUPDQ, and the resulting segments and
#  their SCI values (these are mostly for my reference).
#

DELIM = "-" * 80

# single group integrations fail in the GLS fitting
# so, keep the two method test separate and mark GLS test as
# expected to fail.  Needs fixing, but the fix is not clear
# to me. [KDG - 19 Dec 2018]

dqflags = {
    "GOOD": 0,  # Good pixel.
    "DO_NOT_USE": 2**0,  # Bad pixel. Do not use.
    "SATURATED": 2**1,  # Pixel saturated during exposure.
    "JUMP_DET": 2**2,  # Jump detected during exposure.
    "CHARGELOSS": 2**7,   # Charge migration (was RESERVED_4)
    "NO_GAIN_VALUE": 2**19,  # Gain cannot be measured.
    "UNRELIABLE_SLOPE": 2**24,  # Slope variance large (i.e., noisy pixel).
}

GOOD = dqflags["GOOD"]
DNU = dqflags["DO_NOT_USE"]
SAT = dqflags["SATURATED"]
JUMP = dqflags["JUMP_DET"]
CHRGL = dqflags["CHARGELOSS"]

DEFAULT_OLS = "OLS_C"


# -----------------------------------------------------------------------------
#                           Test Suite
def test_pix_0():
    """
    CASE A: segment has >2 groups, at end of ramp.
    SCI seg is [15., 25., 35., 54., 55., 65., 75., 94., 95., 105.](A)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD] * ngroups
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    # [data, dq, err, var_p, var_r]
    p_true = [1.0117551, GOOD, 0.0921951, 0.0020202, 0.00647973]

    # Set truth values for OPTIONAL results:
    # [slope, sigslope, var_poisson, var_rnoise, yint, sigyint, ped, weights]
    o_true = [1.0117551, 4.874572, 0.0020202, 0.00647973, 15.911023, 27.789335, 13.988245, 13841.038]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_1():
    """
    CASE H: the segment has a good 1st group and a bad 2nd group, so is a
      single group. If there are no later and longer segments in the ramp,
      this group's data will be used in the 'fit'. If there are later and
      longer segments, this group's data will not be used.
    CASE F: segment has 2 good groups not at array end.
    SCI segs are: seg0[15] (H, ignored), seg1[35, 54] (F)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD] * ngroups
    dq[1] = JUMP
    dq[2] = JUMP
    dq[4:] = [SAT] * (ngroups - 4)
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.8999999, JUMP, 1.05057204, 0.03454545, 1.0691562]

    # Set truth values for OPTIONAL results:
    o_true = [1.9, 56.870003, 0.03454545, 1.0691562, -3.0, 56.870003, 13.1, 0.82091206]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_2():
    """
    CASE B: segment has >2 groups, not at end of ramp.
    CASE F: (twice) segment has 2 good groups not at array end.
    SCI segs are: seg0[15,25,35](B), seg1[54,55](F), seg2[65,75](F)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD, GOOD, GOOD, JUMP, GOOD, JUMP, GOOD, JUMP, SAT, SAT]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [0.84833729, JUMP, 0.42747884, 0.00454545, 0.1781927]

    # Set truth values for OPTIONAL results for all segments
    o_true = [
        [1.0000001, 0.1, 1.0],  # slopes for 3 segments
        [28.435, 56.870003, 56.870003],  # sigslope
        [0.00909091, 0.01818182, 0.01818182],  # var_poisson
        [0.26728904, 1.0691562, 1.0691562],  # var_rnoise
        [14.999998, 51.0, 15.0],  # yint
        [36.709427, 56.870003, 56.870003],  # sigyint
        [14.151663],  # pedestal
        [13.091425, 0.84580624, 0.84580624],  # weights
    ]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_3():
    """
    CASE B: segment has >2 groups, not at end of ramp.
    CASE E: segment has 2 good groups, at end of ramp.
    SCI segs are: seg0[15,25,35,54,55,65,75,94](B), seg1[95,105](E)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD] * ngroups
    dq[-2] = JUMP
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.0746869, JUMP, 0.12186482, 0.00227273, 0.01257831]

    # Set truth values for OPTIONAL results:
    o_true = [
        [1.0757396, 1.0],
        [6.450687, 56.870003],
        [0.0025974, 0.01818182],
        [0.01272805, 1.0691562],
        [14.504965, 15.0],
        [27.842508, 56.870003],
        [13.925313],
        [4.2576841e03, 8.458062e-01],
    ]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_4():
    """
    CASE G: segment is the good 1st group of the entire ramp, and no later
      groups are good.
    SCI seg is seg0[15](G)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 1055.0, 1065.0, 1075.0, 2594.0, 2595.0, 2605.0], dtype=np.float32
    )
    dq = [GOOD] + [SAT] * (ngroups - 1)
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.5, GOOD, 1.047105, 0.02727273, 1.0691562]

    # Set truth values for OPTIONAL results:
    o_true = [1.5, 0.0, 0.02727273, 1.0691562, 0.0, 0.0, 13.5, 0.8318386]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


"""
    NOTE:
        There are small differences in the slope computation due to architectural
        differences of C and python.
        Switching to doubles from floats in the C code fixed this problem.

--------------------------------------------------------------------------------
*** [2627] Segment 2, Integration 0 ***
Debug - [slope_fitter.c:3030]     sumx  = 0.018815478310
[ols_fit.py:3818]           sumx = array([0.01881548])

Debug - [slope_fitter.c:3031]     sumxx = 0.132461413741
[ols_fit.py:3819]          sumxx = array([0.13246141])

Debug - [slope_fitter.c:3032]     sumy  = 6.023876190186
[ols_fit.py:3820]           sumy = array([6.0238767], dtype=float32)

Debug - [slope_fitter.c:3033]     sumxy = 39.258270263672
[ols_fit.py:3821]          sumxy = array([39.25826825])

Debug - [slope_fitter.c:3034]     sumw  = 0.002894689096
[ols_fit.py:3822]     nreads_wtd = array([0.00289469], dtype=float32)

================================================================================
================================================================================

    num = (nreads_wtd * sumxy - sumx * sumy)
    denominator = nreads_wtd * sumxx - sumx**2
    invden = 1. / denominator

--------------------------------------------------------------------------------
Debug - [slope_fitter.c:2628] num = 0.000298373401
[ols_fit.py:3281]          num    = 0.000298359918

Debug - [slope_fitter.c:2629] den = 0.000029412389
[ols_fit.py:3282]          den    = 0.000029412383

Debug - [slope_fitter.c:2630] invden = 33999.277343750000
[ols_fit.py:3283]             invden = 33999.284936596494

Debug - [slope_fitter.c:2631] slope = 10.144479751587
[ols_fit.py:3284]            slope  = 10.144023881026

Debug - [slope_fitter.c:2632] gtime = 10.000000000000
Debug - [slope_fitter.c:2633] seg->slope = 1.014447927475
"""


def test_pix_5():
    """
    CASE B: segment has >2 groups, not at end of ramp.
    CASE A: segment has >2 groups, at end of ramp.
    SCI segs are: seg0[15, 25, 35, 54](B), seg1[ 2055, 2065, 2075, 2094, 2095,
       2105](A)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 2055.0, 2065.0, 2075.0, 2094.0, 2095.0, 2105.0], dtype=np.float32
    )
    dq = [GOOD] * ngroups
    dq[4] = JUMP
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.076075, JUMP, 0.16134359, 0.00227273, 0.02375903]

    # Set truth values for OPTIONAL results:
    o_true = [
        [1.2799551, 1.0144024],
        [18.312422, 9.920552],
        [0.00606061, 0.00363636],
        [0.10691562, 0.03054732],
        [13.537246, 2015.0737],
        [35.301933, 67.10882],
        [13.923912],
        [78.34764, 855.78046],
    ]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_6():
    """
    CASE F: segment has 2 good groups not at array end
    CASE A: segment has >2 groups, at end of ramp.
    SCI segs are: seg0[15,25](F), seg1[54, 55, 65, 375, 394, 395, 405](A)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 375.0, 394.0, 395.0, 405.0], dtype=np.float32
    )
    dq = [GOOD] * ngroups
    dq[2] = JUMP
    dq[3] = JUMP
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [6.092052, JUMP, 0.14613187, 0.0025974, 0.01875712]

    # Set truth values for OPTIONAL results:
    o_true = [
        [1.0, 6.195652],
        [56.870003, 8.8936615],
        [0.01818182, 0.0030303],
        [1.0691562, 0.01909207],
        [15.0, -143.2391],
        [56.870003, 58.76999],
        [8.907948],
        [8.4580624e-01, 2.0433204e03],
    ]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_7():
    """
    CASE B: segment has >2 groups, not at end of ramp.
    SCI seg is seg0[15,25,35,54,55,65,75,94](B)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 195.0, 205.0], dtype=np.float32
    )
    dq = [GOOD] * (ngroups - 2) + [JUMP, JUMP]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.0757396, JUMP, 0.12379601, 0.0025974, 0.01272805]

    # Set truth values for OPTIONAL results:
    o_true = [1.0757396, 6.450687, 0.0025974, 0.01272805, 14.504951, 27.842508, 13.92426, 4257.684]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_8():
    """
    CASE H: the segment has a good 1st group and a bad 2nd group.
    CASE B: segment has >2 groups, not at end of ramp.
    SCI segs are: seg0[15](H, ignored), seg1[25, 35, 54, 55, 65, 75](B)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD, JUMP, GOOD, GOOD, GOOD, GOOD, GOOD, SAT, SAT, SAT]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [0.98561335, JUMP, 0.1848883, 0.00363636, 0.03054732]

    # Set truth values for OPTIONAL results:
    o_true = [0.98561335, 9.920554, 0.00363636, 0.03054732, 16.508228, 39.383667, 14.014386, 855.78046]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_9():
    """
    CASE F: segment has 2 good groups not at array end.
    CASE B: segment has >2 groups, not at end of ramp.
    CASE E: segment has 2 good groups, at end of ramp.
    SCI seg are: seg0[15,25](F), seg1[54, 55, 65, 75, 94](B), seg2[95, 105](E)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD, GOOD, JUMP, JUMP, GOOD, GOOD, GOOD, GOOD, JUMP, GOOD]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [0.9999994, JUMP, 0.22721863, 0.0030303, 0.048598]

    # Set truth values for OPTIONAL results:
    o_true = [
        [1.0, 0.9999994, 1.0],
        [56.870003, 13.036095, 56.870003],
        [0.01818182, 0.00454545, 0.01818182],
        [1.0691562, 0.05345781, 1.0691562],
        [15.0, 20.119896, 15.0],
        [56.870003, 68.618195, 56.870003],
        [14.0],
        [0.84580624, 297.23172, 0.84580624],
    ]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_10():
    """
    CASE F: segment has 2 good groups not at array end.
    CASE B: segment has >2 groups, not at end of ramp.
    CASE A: segment has >2 groups, at end of ramp.
    SCI segs are: seg0[15,25](F), seg1[35,54,55](B), seg2[65,75,94,95,105](A)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD, GOOD, JUMP, GOOD, GOOD, JUMP, GOOD, GOOD, GOOD, GOOD]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.0, JUMP, 0.21298744, 0.0025974, 0.04276625]

    # Set truth values for OPTIONAL results:
    o_true = [
        [1.0, 1.0000014, 0.99999964],
        [56.870003, 28.434996, 13.036095],
        [0.01818182, 0.00909091, 0.00454545],
        [1.0691562, 0.26728904, 0.05345781],
        [15.0, 17.999956, 15.000029],
        [56.870003, 88.40799, 93.73906],
        [14.0],
        [0.84580624, 13.091425, 297.23172],
    ]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_11():
    """
    CASE F: segment has 2 good groups not at array end.
    SCI seg is: seg0[15,25](F)
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [15.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 105.0], dtype=np.float32
    )
    dq = [GOOD, GOOD] + [SAT] * (ngroups - 2)
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.0, GOOD, 1.042755, 0.01818182, 1.0691562]

    # Set truth values for OPTIONAL results:
    o_true = [1.0, 56.870003, 0.01818182, 1.0691562, 15.0, 56.870003, 14.0, 0.84580624]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_pix_12():
    """
    CASE NGROUPS=2: the segment has a good 1st group and a saturated 2nd group,
      so is a single group. Group 1's data will be used in the 'fit'.
    """

    # XXX problem with C

    nints, ngroups, nrows, ncols = 1, 2, 1, 2
    gain, rnoise = 5.5, 10.34
    timing = 10.0
    dims = (nints, ngroups, nrows, ncols)
    var = (rnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing)

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array([15.0, 59025.0], dtype=np.float32)
    ramp_data.groupdq[0, :, 0, 0] = np.array([0, SAT])
    ramp_data.data[0, :, 0, 1] = np.array([61000.0, 61000.0], dtype=np.float32)
    ramp_data.groupdq[0, :, 0, 1] = np.array([SAT, SAT])

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results for pixel 1:
    # slope, dq, err, var_p, var_r
    # slope = group1 / deltatime = 15 / 10 = 1.5
    # dq = 2 (saturation) because group2 is saturated, but DNU is *not* set
    p_true = [1.5, GOOD, 1.047105, 0.027273, 1.069156]

    # Set truth values for OPTIONAL results:
    # slope, sig_slope, var_p, var_r, yint, sig_yint, pedestal, weights
    # slope = group1 / deltatime = 15 / 10 = 1.5
    # sig_slope, yint, sig_yint, and pedestal are all zero, because only 1 good group
    o_true = [1.5, 0.0, 0.027273, 1.069156, 0.0, 0.0, 13.5, 0.831839]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)

    # Set truth values for PRIMARY results for pixel 2:
    # slope, dq, err, var_p, var_r
    # slope = zero, because no good data
    # dq = 3 (saturation + do_not_use) because both groups are saturated
    p_true = [np.nan, 3, 0.0, 0.0, 0.0]

    # Set truth values for OPTIONAL results:
    # slope, sig_slope, var_p, var_r, yint, sig_yint, pedestal, weights
    # all values zero, because no good data
    o_true = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    assert_pri(p_true, slopes, 1)
    assert_opt(o_true, ols_opt, 1)


# -------------- start of MIRI tests: all have only a single segment-----
def test_miri_0():
    """
    MIRI data with ramp's 0th and final groups are flagged as DNU
    SCI seg is: [8888., 25., 35., 54., 55., 65., 75., 94., 95., 888.]
    GROUPDQ is: [1, 0, 0, 0, 0, 0, 0, 0, 0, 1]
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing, ts_name="MIRI")

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [8888.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 888.0], dtype=np.float32
    )
    dq = [DNU] + [GOOD] * (ngroups - 2) + [DNU]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.025854, GOOD, 0.12379601, 0.0025974, 0.01272805]

    # Set truth values for OPTIONAL results:
    o_true = [1.025854, 6.450687, 0.0025974, 0.01272805, 26.439266, 27.842508, 23.974146, 4257.684]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_miri_1():
    """
    MIRI data with ramp's 0th and final groups flagged as DNU; 0th group
    is also as a cosmic ray
    SCI seg is: [7777., 125., 135., 154., 165., 175., 185., 204., 205., 777.]
    GROUPDQ is: [5, 0, 0, 0, 0, 0, 0, 0, 0, 1]
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing, ts_name="MIRI")

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [7777.0, 125.0, 135.0, 154.0, 165.0, 175.0, 185.0, 204.0, 205.0, 777.0], dtype=np.float32
    )
    dq = [DNU | JUMP] + [GOOD] * (ngroups - 2) + [DNU]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.1996487, GOOD, 0.12379601, 0.0025974, 0.01272805]

    # Set truth values for OPTIONAL results:
    o_true = [1.1996487, 6.450687, 0.0025974, 0.01272805, 126.110214, 27.842508, 123.800354, 4257.684]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_miri_2():
    """
    MIRI data with ramp's 0th and final groups flagged as both DNU
    and as CR.
    SCI seg is: [4444., 25., 35., 54., 55., 65., 75., 94., 95., 444.]
    GROUPDQ is: [5, 0, 0, 0, 0, 0, 0, 0, 0, 5]
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing, ts_name="MIRI")

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [4444.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 444.0], dtype=np.float32
    )
    dq = [DNU | JUMP] + [GOOD] * (ngroups - 2) + [DNU | JUMP]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.025854, GOOD, 0.12379601, 0.0025974, 0.01272805]

    # Set truth values for OPTIONAL results:
    o_true = [1.025854, 6.450687, 0.0025974, 0.01272805, 26.439266, 27.842508, 23.974146, 4257.684]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


def test_miri_3():
    """
    MIRI data with ramp's 0th and final groups flagged as DNU, and final
    group also flagged as CR.
    SCI seg is: [6666., 25., 35., 54., 55., 65., 75., 94., 95., 666.]
    GROUPDQ is: [1, 0, 0, 0, 0, 0, 0, 0, 0, 5]
    """
    nints, ngroups, nrows, ncols, timing, gain, readnoise = set_scalars()
    dims = (nints, ngroups, nrows, ncols)
    var = (readnoise, gain)
    ramp_data, gain, rnoise = create_blank_ramp_data(dims, var, timing, ts_name="MIRI")

    # Populate pixel-specific SCI and GROUPDQ arrays
    ramp_data.data[0, :, 0, 0] = np.array(
        [6666.0, 25.0, 35.0, 54.0, 55.0, 65.0, 75.0, 94.0, 95.0, 666.0], dtype=np.float32
    )
    dq = [DNU] + [GOOD] * (ngroups - 2) + [DNU | JUMP]
    ramp_data.groupdq[0, :, 0, 0] = np.array(dq)

    save_opt, ncores, bufsize, algo = True, "none", 1024 * 30000, DEFAULT_OLS
    slopes, cube, ols_opt, gls_opt = ramp_fit_data(
        ramp_data, bufsize, save_opt, rnoise, gain, algo, "optimal", ncores, dqflags
    )

    # Set truth values for PRIMARY results:
    p_true = [1.025854, GOOD, 0.12379601, 0.0025974, 0.01272805]

    # Set truth values for OPTIONAL results:
    o_true = [1.025854, 6.450687, 0.0025974, 0.01272805, 26.439266, 27.842508, 23.974146, 4257.684]

    assert_pri(p_true, slopes, 0)
    assert_opt(o_true, ols_opt, 0)


###############################################################################
# utilities


def set_scalars():
    """
    Set needed scalars for the size of the dataset, and other values needed for
    the fit.
    """

    nints = 1
    ngroups = 10
    nrows = 1
    ncols = 1

    timing = 10.0
    gain = 5.5
    readnoise = 10.34

    return nints, ngroups, nrows, ncols, timing, gain, readnoise


def create_blank_ramp_data(dims, var, timing, ts_name="NIRSpec"):
    """
    Create empty RampData classes, as well as gain and read noise arrays,
    based on dimensional, variance, and timing input.
    """
    nints, ngroups, nrows, ncols = dims
    rnval, gval = var
    frame_time = timing
    group_time = timing
    nframes = 1
    groupgap = 0

    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    err = np.ones(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    pixdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint8)
    dark_current = np.zeros(shape=(nrows, ncols), dtype=np.float32)

    ramp_data = RampData()
    ramp_data.set_arrays(data=data, err=err, groupdq=gdq, pixeldq=pixdq, average_dark_current=dark_current)
    ramp_data.set_meta(
        name=ts_name,
        frame_time=frame_time,
        group_time=group_time,
        groupgap=groupgap,
        nframes=nframes,
        drop_frames1=None,
    )
    ramp_data.set_dqflags(dqflags)

    gain = np.ones(shape=(nrows, ncols), dtype=np.float32) * gval
    rnoise = np.ones(shape=(nrows, ncols), dtype=np.float32) * rnval

    return ramp_data, gain, rnoise


def debug_pri(p_true, new_info, pix):
    data, dq, vp, vr, err = new_info

    print(DELIM)
    dbg_print(f"data   = {data[0, pix]}")
    dbg_print(f"p_true = {p_true[0]}")
    print(DELIM)
    dbg_print(f"dq     = {dq[0, pix]}")
    dbg_print(f"p_true = {p_true[1]}")
    print(DELIM)
    dbg_print(f"vp     = {vp[0, pix]}")
    dbg_print(f"p_true = {p_true[3]}")
    print(DELIM)
    dbg_print(f"vr     = {vr[0, pix]}")
    dbg_print(f"p_true = {p_true[4]}")
    print(DELIM)
    dbg_print(f"err    = {err[0, pix]}")
    dbg_print(f"p_true = {p_true[2]}")
    print(DELIM)


def assert_pri(p_true, new_info, pix):
    """
    Compare true and fit values of primary output for extensions
    SCI, DQ, ERR, VAR_POISSON, VAR_RNOISE.
    """

    data, dq, var_poisson, var_rnoise, err = new_info

    npt.assert_allclose(data[0, pix], p_true[0], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(dq[0, pix], p_true[1], atol=1e-1)
    npt.assert_allclose(err[0, pix], p_true[2], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(var_poisson[0, pix], p_true[3], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(var_rnoise[0, pix], p_true[4], atol=2e-5, rtol=2e-5)


def debug_opt(o_true, opt_info, pix):
    (slope, sigslope, var_poisson, var_rnoise, yint, sigyint, pedestal, weights, crmag) = opt_info

    opt_slope = slope[0, :, 0, pix]
    opt_sigslope = sigslope[0, :, 0, pix]
    opt_var_poisson = var_poisson[0, :, 0, pix]
    opt_var_rnoise = var_rnoise[0, :, 0, pix]
    opt_yint = yint[0, :, 0, pix]
    opt_sigyint = sigyint[0, :, 0, pix]
    opt_pedestal = pedestal[:, 0, pix]
    opt_weights = weights[0, :, 0, pix]

    print(DELIM)
    dbg_print(f"slope  = {opt_slope}")
    dbg_print(f"o_true = {o_true[0]}")
    print(DELIM)
    dbg_print(f"sigslope = {opt_sigslope}")
    dbg_print(f"o_true   = {o_true[1]}")
    print(DELIM)
    dbg_print(f"var_p  = {opt_var_poisson}")
    dbg_print(f"o_true = {o_true[2]}")
    print(DELIM)
    dbg_print(f"var_r  = {opt_var_rnoise}")
    dbg_print(f"o_true = {o_true[3]}")
    print(DELIM)
    dbg_print(f"yint   = {opt_yint}")
    dbg_print(f"o_true = {o_true[4]}")
    print(DELIM)
    dbg_print(f"sigyint = {opt_sigyint}")
    dbg_print(f"o_true  = {o_true[5]}")
    print(DELIM)
    dbg_print(f"pedestal = {opt_pedestal}")
    dbg_print(f"o_true   = {o_true[6]}")
    print(DELIM)
    dbg_print(f"weights = {opt_weights}")
    dbg_print(f"o_true  = {o_true[7]}")
    print(DELIM)


def assert_opt(o_true, opt_info, pix):
    """
    Compare true and fit values of optional output for extensions SLOPE,
    SIGSLOPE, VAR_POISSON, VAR_RNOISE, YINT, SIGYINT, PEDESTAL, and WEIGHTS.
    Selecting the particular (and only) ramp in the optional output, which is
    [0,:,0,0]
    """
    (slope, sigslope, var_poisson, var_rnoise, yint, sigyint, pedestal, weights, crmag) = opt_info

    opt_slope = slope[0, :, 0, pix]
    opt_sigslope = sigslope[0, :, 0, pix]
    opt_var_poisson = var_poisson[0, :, 0, pix]
    opt_var_rnoise = var_rnoise[0, :, 0, pix]
    opt_yint = yint[0, :, 0, pix]
    opt_sigyint = sigyint[0, :, 0, pix]
    opt_pedestal = pedestal[:, 0, pix]
    opt_weights = weights[0, :, 0, pix]

    npt.assert_allclose(opt_slope, o_true[0], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(opt_sigslope, o_true[1], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(opt_var_poisson, o_true[2], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(opt_var_rnoise, o_true[3], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(opt_yint, o_true[4], atol=2e-2)
    npt.assert_allclose(opt_sigyint, o_true[5], atol=2e-5, rtol=2e-5)
    npt.assert_allclose(opt_pedestal, o_true[6], atol=2e-5, rtol=3e-5)
    npt.assert_allclose(opt_weights, o_true[7], atol=2e-5, rtol=2e-5)


def dbg_print(string):
    """
    Print string with line number and filename.
    """
    cf = inspect.currentframe()
    line_number = cf.f_back.f_lineno
    finfo = inspect.getframeinfo(cf.f_back)
    fname = Path(finfo.filename).name
    print(f"[{fname}:{line_number}] {string}")



import gwcs
import numpy as np
import pytest
from astropy import coordinates as coord
from astropy import units as u
from astropy import wcs as fitswcs
from astropy.io import fits
from astropy.modeling import models
from gwcs import coordinate_frames as cf

from stcal.alignment import resample_utils
from stcal.alignment.util import (
    _validate_wcs_list,
    compute_fiducial,
    _compute_fiducial_from_footprints,
    compute_s_region_imaging,
    compute_s_region_keyword,
    compute_scale,
    reproject,
    _sregion_to_footprint,
    wcs_bbox_from_shape,
    wcs_from_footprints,
    wcs_from_sregions
)


def _create_wcs_object_without_distortion(
    fiducial_world,
    pscale,
    shape,
):
    # subtract 1 to account for pixel indexing starting at 0
    shift = models.Shift() & models.Shift()

    scale = models.Scale(pscale[0]) & models.Scale(pscale[1])

    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        fiducial_world[0],
        fiducial_world[1],
        180,
    )

    det2sky = shift | scale | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix))
    sky_frame = cf.CelestialFrame(reference_frame=coord.FK5(), name="fk5", unit=(u.deg, u.deg))

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]

    wcs_obj = gwcs.WCS(pipeline)

    wcs_obj.bounding_box = (
        (-0.5, shape[-1] - 0.5),
        (-0.5, shape[-2] - 0.5),
    )

    return wcs_obj


def _create_wcs_and_datamodel(fiducial_world, shape, pscale):
    wcs = _create_wcs_object_without_distortion(fiducial_world=fiducial_world, shape=shape, pscale=pscale)
    ra_ref, dec_ref = fiducial_world[0], fiducial_world[1]
    return DataModel(
        ra_ref=ra_ref,
        dec_ref=dec_ref,
        roll_ref=0,
        v2_ref=0,
        v3_ref=0,
        v3yangle=0,
        wcs=wcs,
    )


class WcsInfo:
    """
    JWST-like wcsinfo object
    """
    def __init__(self, ra_ref, dec_ref, roll_ref, v2_ref, v3_ref, v3yangle):
        self.ra_ref = ra_ref
        self.dec_ref = dec_ref
        self.ctype1 = "RA---TAN"
        self.ctype2 = "DEC--TAN"
        self.v2_ref = v2_ref
        self.v3_ref = v3_ref
        self.v3yangle = v3yangle
        self.roll_ref = roll_ref
        self.vparity = -1
        self.wcsaxes = 2
        self.s_region = ""
        self.instance = self.instance()

    def instance(self):
        return {
            "ra_ref": self.ra_ref,
            "dec_ref": self.dec_ref,
            "ctype1": self.ctype1,
            "ctype2": self.ctype2,
            "v2_ref": self.v2_ref,
            "v3_ref": self.v3_ref,
            "v3yangle": self.v3yangle,
            "roll_ref": self.roll_ref,
            "vparity": self.vparity,
            "wcsaxes": self.wcsaxes,
            "s_region": self.s_region,
        }


class Coordinates:
    def __init__(self):
        self.reference_frame = "ICRS"


class MetaData:
    def __init__(self, ra_ref, dec_ref, roll_ref, v2_ref, v3_ref, v3yangle, wcs=None):
        self.wcsinfo = WcsInfo(ra_ref, dec_ref, roll_ref, v2_ref, v3_ref, v3yangle)
        self.wcs = wcs
        self.coordinates = Coordinates()


class DataModel:
    """JWST-like datamodel object"""
    def __init__(self, ra_ref, dec_ref, roll_ref, v2_ref, v3_ref, v3yangle, wcs=None):
        self.meta = MetaData(ra_ref, dec_ref, roll_ref, v2_ref, v3_ref, v3yangle, wcs=wcs)


@pytest.mark.parametrize("footprint", [True, False])
def test_compute_fiducial(footprint):
    """Test that util.compute_fiducial can properly determine the center of the
    WCS's footprint.
    """

    shape = (3, 3)  # in pixels
    fiducial_world = (0, 0)  # in deg
    pscale = (0.000014, 0.000014)  # in deg/pixel

    wcs = _create_wcs_object_without_distortion(fiducial_world=fiducial_world, shape=shape, pscale=pscale)
    if footprint:
        footprint = wcs.footprint()
        computed_fiducial = _compute_fiducial_from_footprints([footprint])
    else:
        computed_fiducial = compute_fiducial([wcs])

    assert all(np.isclose(wcs(1, 1), computed_fiducial))


@pytest.mark.parametrize("pscales", [(0.000014, 0.000014), (0.000028, 0.000014)])
def test_compute_scale(pscales):
    """Test that util.compute_scale can properly determine the pixel scale of a
    WCS object.
    """
    shape = (3, 3)  # in pixels
    fiducial_world = (0, 0)  # in deg
    pscale = (pscales[0], pscales[1])  # in deg/pixel

    wcs = _create_wcs_object_without_distortion(fiducial_world=fiducial_world, shape=shape, pscale=pscale)
    expected_scale = np.sqrt(pscale[0] * pscale[1])

    computed_scale = compute_scale(wcs=wcs, fiducial=fiducial_world)

    assert np.isclose(expected_scale, computed_scale)


def test_sregion_to_footprint():
    """Test that util._sregion_to_footprint can properly convert an S_REGION
    string to a list of vertices.
    """
    s_region = "POLYGON ICRS  1.000000000 2.000000000 3.000000000 4.000000000 5.000000000 6.000000000 7.000000000 8.000000000"
    expected_footprint = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])

    footprint = _sregion_to_footprint(s_region)

    assert footprint.shape == (4,2)
    assert np.allclose(footprint, expected_footprint)


@pytest.mark.parametrize("s_regions", [True, False])
def test_wcs_from_footprints(s_regions):
    """
    Test that the WCS created from wcs_from_footprints has correct vertice coordinates.

    N.B.: this test will create two 3x3 arrays shifted by 0.000028 deg in
    both directions, which means that the combined WCS generated by wcs_from_footprints
    should be a 4x4 array with its fiducial point coordinates equal to the
    first element of its footprint.
    """
    shape = (3, 3)  # in pixels
    fiducial_world = (10, 0)  # in deg
    pscale = (0.000028, 0.000028)  # in deg/pixel
    dm_1 = _create_wcs_and_datamodel(fiducial_world, shape, pscale)
    wcs_1 = dm_1.meta.wcs

    # shift fiducial by the size of a pixel projected onto the sky in both directions
    # and create a new WCS
    fiducial_world = (
        fiducial_world[0] - 0.000028,
        fiducial_world[1] - 0.000028,
    )
    dm_2 = _create_wcs_and_datamodel(fiducial_world, shape, pscale)
    wcs_2 = dm_2.meta.wcs
    if s_regions:
        footprints = [wcs_1.footprint(), wcs_2.footprint()]
        wcs = wcs_from_sregions(footprints, wcs_1, dm_1.meta.wcsinfo.instance)
    else:
        wcs_list = [wcs_1, wcs_2]
        wcs = wcs_from_footprints(wcs_list, wcs_1, dm_1.meta.wcsinfo.instance)

    # check that all elements of footprint match the *vertices* of the new combined WCS
    assert all(np.isclose(wcs.footprint()[0], wcs(0, 0)))
    assert all(np.isclose(wcs.footprint()[1], wcs(0, 4)))
    assert all(np.isclose(wcs.footprint()[2], wcs(4, 4)))
    assert all(np.isclose(wcs.footprint()[3], wcs(4, 0)))

    # check that fiducials match their expected coords in the new combined WCS
    assert all(np.isclose(wcs_1(0, 0), wcs(2.5, 1.5)))
    assert all(np.isclose(wcs_2(0, 0), wcs(3.5, 0.5)))


def test_validate_wcs_list():
    shape = (3, 3)  # in pixels
    fiducial_world = (10, 0)  # in deg
    pscale = (0.000028, 0.000028)  # in deg/pixel

    dm_1 = _create_wcs_and_datamodel(fiducial_world, shape, pscale)
    wcs_1 = dm_1.meta.wcs

    # shift fiducial by one pixel in both directions and create a new WCS
    fiducial_world = (
        fiducial_world[0] - 0.000028,
        fiducial_world[1] - 0.000028,
    )
    dm_2 = _create_wcs_and_datamodel(fiducial_world, shape, pscale)
    wcs_2 = dm_2.meta.wcs

    wcs_list = [wcs_1, wcs_2]

    assert _validate_wcs_list(wcs_list)


@pytest.mark.parametrize(
    ("wcs_list", "expected_error"),
    [
        ([], TypeError),
        ([1, 2, 3], TypeError),
        (["1", "2", "3"], TypeError),
        (["1", None, []], TypeError),
        ("1", TypeError),
        (1, ValueError),
        (None, ValueError),
    ],
)
def test_validate_wcs_list_invalid(wcs_list, expected_error):
    with pytest.raises(expected_error, match=r".*"):
        _validate_wcs_list(wcs_list)


def get_fake_wcs():
    fake_wcs1 = fitswcs.WCS(
        fits.Header(
            {
                "NAXIS": 2,
                "NAXIS1": 4,
                "NAXIS2": 4,
                "CTYPE1": "RA---TAN",
                "CTYPE2": "DEC--TAN",
                "CRVAL1": 0,
                "CRVAL2": 0,
                "CRPIX1": 1,
                "CRPIX2": 1,
                "CDELT1": -0.1,
                "CDELT2": 0.1,
            }
        )
    )
    fake_wcs2 = fitswcs.WCS(
        fits.Header(
            {
                "NAXIS": 2,
                "NAXIS1": 5,
                "NAXIS2": 5,
                "CTYPE1": "RA---TAN",
                "CTYPE2": "DEC--TAN",
                "CRVAL1": 0,
                "CRVAL2": 0,
                "CRPIX1": 1,
                "CRPIX2": 1,
                "CDELT1": -0.05,
                "CDELT2": 0.05,
            }
        )
    )
    return fake_wcs1, fake_wcs2


@pytest.mark.parametrize(
    ("x_inp", "y_inp", "x_expected", "y_expected"),
    [
        (1000, 2000, np.array(2000), np.array(4000)),  # string input test
        ([1000], [2000], np.array(2000), np.array(4000)),  # array input test
        pytest.param(1, 2, 3, 4, marks=pytest.mark.xfail),  # expected failure test
    ],
)
def test_reproject(x_inp, y_inp, x_expected, y_expected):
    wcs1, wcs2 = get_fake_wcs()
    f = reproject(wcs1, wcs2)
    x_out, y_out = f(x_inp, y_inp)
    assert np.allclose(x_out, x_expected, rtol=1e-05)
    assert np.allclose(y_out, y_expected, rtol=1e-05)


def test_wcs_bbox_from_shape_2d():
    bb = wcs_bbox_from_shape((512, 2048))
    assert bb == ((-0.5, 2047.5), (-0.5, 511.5))


def test_wcs_bbox_from_shape_3d():
    bb = wcs_bbox_from_shape((3, 32, 2048))
    assert bb == ((-0.5, 2047.5), (-0.5, 31.5))

    bb = wcs_bbox_from_shape((750, 45, 50))
    assert bb == ((-0.5, 49.5), (-0.5, 44.5))


@pytest.mark.parametrize(
    ("shape", "pixmap_expected_shape"),
    [
        (None, (4, 4, 2)),
        ((100, 200), (100, 200, 2)),
    ],
)
def test_calc_pixmap_shape(shape, pixmap_expected_shape):
    # TODO: add test for gwcs.WCS
    wcs1, wcs2 = get_fake_wcs()
    pixmap = resample_utils.calc_pixmap(wcs1, wcs2, shape=shape)
    assert pixmap.shape == pixmap_expected_shape


@pytest.mark.parametrize(
    ("model", "footprint", "expected_s_region", "expected_log_info"),
    [
        (
            _create_wcs_and_datamodel((10, 0), (3, 3), (0.000028, 0.000028)),
            np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]),
            "POLYGON ICRS  1.000000000 2.000000000 3.000000000 4.000000000 5.000000000 6.000000000 7.000000000 8.000000000",  # noqa: E501
            "Update S_REGION to POLYGON ICRS  1.000000000 2.000000000 3.000000000 4.000000000 5.000000000 6.000000000 7.000000000 8.000000000",  # noqa: E501
        ),
        (
            _create_wcs_and_datamodel((10, 0), (3, 3), (0.000028, 0.000028)),
            np.array([[1.0, 2.0], [3.0, np.nan], [5.0, 6.0], [7.0, 8.0]]),
        None,
            "There are NaNs in s_region, S_REGION not updated.",
        ),
    ],
)
def test_compute_s_region_keyword(model, footprint, expected_s_region, expected_log_info, caplog):
    """
    Test that S_REGION keyword is being properly populated with the coordinate values.
    """
    model.meta.wcsinfo.s_region = compute_s_region_keyword(footprint)
    assert model.meta.wcsinfo.s_region == expected_s_region
    assert expected_log_info in caplog.text


@pytest.mark.parametrize(
    ("shape", "expected_bbox"),
    [
        ((100, 200), ((-0.5, 199.5), (-0.5, 99.5))),
        ((1, 1), ((-0.5, 0.5), (-0.5, 0.5))),
        ((0, 0), ((-0.5, -0.5), (-0.5, -0.5))),
    ],
)
def test_wcs_bbox_from_shape(shape, expected_bbox):
    """
    Test that the bounding box generated by wcs_bbox_from_shape is correct.
    """
    bbox = wcs_bbox_from_shape(shape)
    assert bbox == expected_bbox


@pytest.mark.parametrize(
    ("model", "bounding_box", "data"),
    [
        (
            _create_wcs_and_datamodel((10, 0), (3, 3), (0.000028, 0.000028)),
            ((-0.5, 2.5), (-0.5, 2.5)),
            None,
        ),
        (
            _create_wcs_and_datamodel((10, 0), (3, 3), (0.000028, 0.000028)),
            None,
            np.zeros((3, 3)),
        ),
    ],
)
def test_compute_s_region_imaging(model, bounding_box, data):
    """
    Test that S_REGION keyword is being properly updated with the coordinates
    corresponding to the footprint (same as WCS(bounding box)).
    """
    model.data = data
    model.meta.wcs.bounding_box = bounding_box
    expected_s_region_coords = [
        *model.meta.wcs(-0.5, -0.5),
        *model.meta.wcs(-0.5, 2.5),
        *model.meta.wcs(2.5, 2.5),
        *model.meta.wcs(2.5, -0.5),
    ]
    shape = data.shape if data is not None else None
    model.meta.wcsinfo.s_region = compute_s_region_imaging(model.meta.wcs, shape=shape, center=False)
    updated_s_region_coords = [float(x) for x in model.meta.wcsinfo.s_region.split(" ")[3:]]
    assert all(np.isclose(x, y) for x, y in zip(updated_s_region_coords,
                                                expected_s_region_coords,
                                                strict=False))



from __future__ import annotations

import math
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import astropy.units as u

if TYPE_CHECKING:
    import astropy
    import gwcs

from astropy.coordinates import SkyCoord
from astropy.table import Table
from tweakwcs.correctors import JWSTWCSCorrector
from tweakwcs.imalign import align_wcs
from tweakwcs.matchutils import XYXYMatch

from stcal.alignment import wcs_from_sregions

from .astrometric_utils import create_astrometric_catalog

_SQRT2 = math.sqrt(2.0)
SINGLE_GROUP_REFCAT = ["GAIADR3", "GAIADR2", "GAIADR1"]
_SINGLE_GROUP_REFCAT_STR = ",".join(SINGLE_GROUP_REFCAT)

__all__ = ["relative_align", "absolute_align", "SINGLE_GROUP_REFCAT",
           "filter_catalog_by_bounding_box"]


class TweakregError(BaseException):
    pass


def relative_align(correctors: list,
                   searchrad: float = 2.0,
                   separation: float = 1.0,
                   use2dhist: bool = True,
                   tolerance: float = 0.7,
                   xoffset: float = 0.0,
                   yoffset: float = 0.0,
                   enforce_user_order: bool = False,
                   expand_refcat: bool = False,
                   minobj: int = 15,
                   fitgeometry: str = "rshift",
                   nclip: int = 3,
                   sigma: float = 3.0,
                   clip_accum: bool = False,
                   ) -> list:

    if separation <= _SQRT2 * tolerance:
        msg = ("Parameter 'separation' must be larger than 'tolerance' by at "
                "least a factor of sqrt(2) to avoid source confusion.")
        raise TweakregError(msg)

    # align images:
    xyxymatch = XYXYMatch(
        searchrad=searchrad,
        separation=separation,
        use2dhist=use2dhist,
        tolerance=tolerance,
        xoffset=xoffset,
        yoffset=yoffset
    )

    try:
        align_wcs(
            correctors,
            refcat=None,
            enforce_user_order=enforce_user_order,
            expand_refcat=expand_refcat,
            minobj=minobj,
            match=xyxymatch,
            fitgeom=fitgeometry,
            nclip=nclip,
            sigma=(sigma, "rmse"),
            clip_accum=clip_accum,
        )

    except ValueError as e:
        msg = e.args[0]
        if (msg == "Too few input images (or groups of images) with "
                "non-empty catalogs."):
            msg += ("At least two exposures are required for "
                    "image alignment. Nothing to do.")
            raise TweakregError(msg) from None
        raise

    except RuntimeError as e:
        msg = e.args[0]
        if msg.startswith("Number of output coordinates exceeded allocation"):
            # we need at least two exposures to perform image alignment
            msg += "Multiple sources within specified tolerance \
                    matched to a single reference source. Try to \
                    adjust 'tolerance' and/or 'separation' parameters."
            raise TweakregError(msg) from None
        raise

    with warnings.catch_warnings(record=True) as w:
        is_small = _is_wcs_correction_small(correctors,
                                            use2dhist,
                                            searchrad,
                                            tolerance,
                                            xoffset,
                                            yoffset)
        msg = "".join([str(mess.message) for mess in w])
        if not is_small:
            msg += " Skipping relative alignment..."
            raise TweakregError(msg)

    return correctors


def absolute_align(correctors: list,
                   abs_refcat: str,
                   ref_wcs: gwcs.wcs.WCS,
                   ref_wcsinfo: dict,
                   epoch: str | astropy.time.Time,
                   save_abs_catalog: bool = False,
                   abs_catalog_output_dir: str | None = None,
                   abs_searchrad: float = 6.0,
                   abs_separation: float = 1.0,
                   abs_use2dhist: bool = True,
                   abs_tolerance: float = 0.7,
                   abs_minobj: int = 15,
                   abs_fitgeometry: str = "rshift",
                   abs_nclip: int = 3,
                   abs_sigma: float = 3.0,
                   clip_accum: bool = False,) -> list:

    if abs_separation <= _SQRT2 * abs_tolerance:
        msg = ("Parameter 'abs_separation' must be larger than "
               "'abs_tolerance' "
               "by at least a factor of sqrt(2) to avoid source confusion.")
        raise TweakregError(msg)

    ref_cat = _parse_refcat(abs_refcat,
                            correctors,
                            ref_wcs,
                            ref_wcsinfo,
                            epoch,
                            save_abs_catalog=save_abs_catalog,
                            output_dir=abs_catalog_output_dir)

    # Check that there are enough GAIA sources for a reliable/valid fit
    num_ref = len(ref_cat)
    if num_ref < abs_minobj:
        msg = (f"Not enough sources ({num_ref}) in the reference catalog "
            "for the single-group alignment step to perform a fit. "
            "Skipping alignment to the input reference catalog!")
        raise TweakregError(msg)

    # align images:
    # Update to separation needed to prevent confusion of sources
    # from overlapping images where centering is not consistent or
    # for the possibility that errors still exist in relative overlap.
    xyxymatch_gaia = XYXYMatch(
        searchrad=abs_searchrad,
        separation=abs_separation,
        use2dhist=abs_use2dhist,
        tolerance=abs_tolerance,
        xoffset=0.0,
        yoffset=0.0
    )

    # Set group_id to same value so all get fit as one observation
    # The assigned value, 987654, has been hard-coded to make it
    # easy to recognize when alignment to GAIA was being performed
    # as opposed to the group_id values used for relative alignment
    # earlier in this step.
    for corrector in correctors:
        corrector.meta["group_id"] = 987654
        if ("fit_info" in corrector.meta and
                "REFERENCE" in corrector.meta["fit_info"]["status"]):
            del corrector.meta["fit_info"]

    # Perform fit
    try:
        align_wcs(
            correctors,
            refcat=ref_cat,
            enforce_user_order=True,
            expand_refcat=False,
            minobj=abs_minobj,
            match=xyxymatch_gaia,
            fitgeom=abs_fitgeometry,
            nclip=abs_nclip,
            sigma=(abs_sigma, "rmse"),
            clip_accum=clip_accum,
        )
    except ValueError as e:
        msg = e.args[0]
        if (msg == "Too few input images (or groups of images) with "
                "non-empty catalogs."):
            msg += ("At least one exposure is required to align images "
                    "to an absolute reference catalog. Alignment to an "
                    "absolute reference catalog will not be performed.")
            raise TweakregError(msg) from None
        raise

    except RuntimeError as e:
        msg = e.args[0]
        if msg.startswith("Number of output coordinates exceeded allocation"):
            # we need at least two exposures to perform image alignment
            msg += ("Multiple sources within specified tolerance "
                    "matched to a single reference source. Try to "
                    "adjust 'tolerance' and/or 'separation' parameters. "
                    "Alignment to an absolute reference catalog will "
                    "not be performed.")
            raise TweakregError(msg) from None
        raise

    return correctors


def _parse_refcat(abs_refcat: str | Path,
                  correctors: list,
                  wcs: gwcs.wcs.WCS,
                  wcsinfo: dict,
                  epoch: str | astropy.time.Time,
                  save_abs_catalog: bool = False,
                  output_dir: str | None = None) -> Table:
    """
    Figure out if abs_refcat is an input filename or
    the name of a GAIA catalog. If the former, load it,
    and if the latter, retrieve that catalog from the Web.
    If desired, save the reference catalog in the specified directory.
    """
    abs_refcat = str(abs_refcat).strip()
    if save_abs_catalog:
        root = f"fit_{abs_refcat.lower()}_ref.ecsv"
        output_name = Path(root) if output_dir is None \
            else Path(output_dir) / root
    else:
        output_name = None

    gaia_cat_name = abs_refcat.upper()
    if gaia_cat_name in SINGLE_GROUP_REFCAT:

        # combine all aligned wcs to compute a new footprint to
        # filter the absolute catalog sources
        combined_wcs = wcs_from_sregions(
                        [corrector.wcs.footprint() for corrector in correctors],
                        ref_wcs=wcs,
                        ref_wcsinfo=wcsinfo,
        )

        return create_astrometric_catalog(
            combined_wcs, epoch,
            catalog=gaia_cat_name,
            output=output_name,
        )

    if Path(abs_refcat).is_file():
        return Table.read(abs_refcat)

    msg = (f"Invalid 'abs_refcat' value: {abs_refcat}. 'abs_refcat' must be "
           "a path to an existing file name or one of the supported "
           f"reference catalogs: {_SINGLE_GROUP_REFCAT_STR}.")
    raise ValueError(msg)


def _is_wcs_correction_small(correctors,
                             use2dhist=True,
                             searchrad=2.0,
                             tolerance=0.7,
                             xoffset=0.0,
                             yoffset=0.0):
    # check for a small wcs correction, it should be small
    if use2dhist:
        max_corr = 2 * (searchrad + tolerance) * u.arcsec
    else:
        max_corr = 2 * (max(abs(xoffset), abs(yoffset)) +
                        tolerance) * u.arcsec
    for corrector in correctors:
        aligned_skycoord = _wcs_to_skycoord(corrector.wcs)
        original_skycoord = corrector.meta["original_skycoord"]
        separation = original_skycoord.separation(aligned_skycoord)
        if not (separation < max_corr).all():
            # Large corrections are typically a result of source
            # mis-matching or poorly-conditioned fit. Skip such models.
            msg = f"WCS has been tweaked by more than {10 * tolerance} arcsec"
            warnings.warn(msg, stacklevel=1)
            return False
    return True


def _wcs_to_skycoord(wcs):
    ra, dec = wcs.footprint(axis_type="spatial").T
    return SkyCoord(ra=ra, dec=dec, unit="deg")


def filter_catalog_by_bounding_box(catalog: Table, bounding_box: list[tuple]) -> Table:
    """
    Given a catalog of x,y positions, only return sources that fall
    inside the bounding box.
    """
    if bounding_box is None:
        return catalog

    # filter out sources outside the WCS bounding box
    ((xmin, xmax), (ymin, ymax)) = bounding_box
    x = catalog["x"]
    y = catalog["y"]
    mask = (x > xmin) & (x < xmax) & (y > ymin) & (y < ymax)
    return catalog[mask]


def construct_wcs_corrector(wcs: gwcs.WCS,
                            refang: dict,
                            catalog: Table,
                            group_id: str,) -> JWSTWCSCorrector:
    """
    pre-compute skycoord here so we can later use it
    to check for a small wcs correction.

    Parameters
    ----------
    wcs : `gwcs.WCS`
        WCS object to be corrected.

    refang : dict
        Dictionary containing WCSreference angles.
    """
    catalog = filter_catalog_by_bounding_box(
        catalog, wcs.bounding_box)

    return JWSTWCSCorrector(
        wcs=wcs,
        wcsinfo={"roll_ref": refang["roll_ref"],
                 "v2_ref": refang["v2_ref"],
                 "v3_ref": refang["v3_ref"]},
        # catalog and group_id are required meta
        meta={
            "catalog": catalog,
            "name": catalog.meta.get("name"),
            "group_id": group_id,
            "original_skycoord": _wcs_to_skycoord(wcs),
        }
    )



# cython: language_level=3str

from libcpp.vector cimport vector


cdef struct RampIndex:
    int start
    int end


cdef struct RampFit:
    float slope
    float read_var
    float poisson_var


ctypedef vector[RampIndex] RampQueue


cdef class ReadPattern:
    cdef float[::1] t_bar
    cdef float[::1] tau
    cdef int[::1] n_reads


cpdef RampQueue init_ramps(int[:] dq,
                           int n_resultants)


cpdef ReadPattern from_read_pattern(list[list[int]] read_pattern,
                                    float read_time,
                                    int n_resultants)


cdef RampFit fit_ramp(float[:] resultants_,
                      float[:] t_bar_,
                      float[:] tau_,
                      int[:] n_reads,
                      float read_noise,
                      RampIndex ramp)



import os

import requests
from astropy import table
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table

from stcal.alignment import compute_fiducial

ASTROMETRIC_CAT_ENVVAR = "ASTROMETRIC_CATALOG_URL"
DEF_CAT_URL = "http://gsss.stsci.edu/webservices"

SERVICELOCATION = os.environ.get(ASTROMETRIC_CAT_ENVVAR, DEF_CAT_URL)

TIMEOUT = 30.0  # in seconds

"""

Primary function for creating an astrometric reference catalog.

"""


__all__ = [
    "TIMEOUT",
    "compute_radius",
    "create_astrometric_catalog",
    "get_catalog"]


def create_astrometric_catalog(
        wcs,
        epoch,
        catalog="GAIADR3",
        output="ref_cat.ecsv",
        gaia_only=False,
        table_format="ascii.ecsv",
        num_sources=None):
    """Create an astrometric catalog that covers the inputs' field-of-view.

    Parameters
    ----------
    wcs : `~astropy.wcs.WCS`
        WCS object specified by the user as generated by
        `resample.resample_utils.make_output_wcs`. This will typically
        have the same plate-scale and orientation as the first member in the
        list of input images to make_output_wcs. Fortunately, for alignment,
        this doesn't matter since no resampling of data will be performed.

    epoch : float
        Reference epoch used to update the coordinates for proper motion
        (in decimal year).

    catalog : str, optional
        Name of catalog to extract astrometric positions for sources in the
        input images' field-of-view. Default: GAIADR3. Options available are
        documented on the catalog web page.

    output : str, optional
        Filename to give to the astrometric catalog read in from the master
        catalog web service.  If None, no file will be written out.

    gaia_only : bool, optional
        Specify whether or not to only use sources from GAIA in output catalog

    num_sources : int
        Maximum number of brightest/faintest sources to return in catalog.
        If `num_sources` is negative, return that number of the faintest
        sources.  By default, all sources are returned.



    Notes
    -----
    This function will point to astrometric catalog web service defined
    through the use of the ASTROMETRIC_CATALOG_URL environment variable.

    Returns
    -------
    ref_table : `~astropy.table.Table`
        Astropy Table object of the catalog
    """
    # start by creating a composite field-of-view for all inputs
    radius, fiducial = compute_radius(wcs)

    # perform query for this field-of-view
    ref_dict = get_catalog(
        fiducial[0],
        fiducial[1],
        epoch=epoch,
        search_radius=radius,
        catalog=catalog)
    if len(ref_dict) == 0:
        return ref_dict

    colnames = ("ra", "dec", "mag", "objID", "epoch")
    ref_table = ref_dict[colnames]

    # Add catalog name as meta data
    ref_table.meta["catalog"] = catalog
    ref_table.meta["gaia_only"] = gaia_only

    # rename coordinate columns to be consistent with tweakwcs
    ref_table.rename_column("ra", "RA")
    ref_table.rename_column("dec", "DEC")

    # Append GAIA ID as a new column to the table...
    gaia_sources = []
    for source in ref_dict:
        if "GAIAsourceID" in source:
            g = source["GAIAsourceID"]
            if gaia_only and g.strip() == "":
                continue
        else:
            g = "-1"  # indicator for no source ID extracted
        gaia_sources.append(g)

    gaia_col = table.Column(data=gaia_sources, name="GaiaID", dtype="U25")
    ref_table.add_column(gaia_col)

    # sort table by magnitude, fainter to brightest
    ref_table.sort("mag", reverse=True)

    # If specified by the use through the 'num_sources' parameter,
    # trim the returned catalog down to the brightest 'num_sources' sources
    # Should 'num_sources' be a negative value, it will return the faintest
    # 'num_sources' sources.
    if num_sources is not None:
        indx = -1 * num_sources
        ref_table = ref_table[:indx] if num_sources < 0 else ref_table[indx:]

    # Write out table to a file, if specified
    if output is not None:
        ref_table.write(output, format=table_format, overwrite=True)

    return ref_table


def compute_radius(wcs):
    """Compute the radius from the center to the furthest edge of the WCS."""
    fiducial = compute_fiducial([wcs,])
    img_center = SkyCoord(
        ra=fiducial[0] * u.degree,
        dec=fiducial[1] * u.degree)
    wcs_foot = wcs.footprint()
    img_corners = SkyCoord(ra=wcs_foot[:, 0] * u.degree,
                           dec=wcs_foot[:, 1] * u.degree)
    radius = img_center.separation(img_corners).max().value

    return radius, fiducial


def get_catalog(
    right_ascension,
    declination,
    epoch=2016.0,
    search_radius=0.1,
    catalog="GAIADR3",
    timeout=TIMEOUT,
):
    """Extract catalog from VO web service.

    Parameters
    ----------
    right_ascension : float
        Right Ascension (RA) of center of field-of-view (in decimal degrees)

    declination : float
        Declination (Dec) of center of field-of-view (in decimal degrees)

    epoch : float, optional
        Reference epoch used to update the coordinates for proper motion
        (in decimal year). Default: 2016.0

    search_radius : float, optional
        Search radius (in decimal degrees) from field-of-view center to use
        for sources from catalog.  Default: 0.1 degrees

    catalog : str, optional
        Name of catalog to query, as defined by web-service. Default: 'GAIADR3'

    timeout : float, optional
        Timeout in seconds to wait for the catalog web service to respond. Default: 30.0 s

    Returns
    -------
    csv : `~astropy.table.Table`
        CSV object of returned sources with all columns as provided by catalog

    """
    service_type = "vo/CatalogSearch.aspx"
    spec_str = "RA={}&DEC={}&EPOCH={}&SR={}&FORMAT={}&CAT={}&MINDET=5"
    headers = {"Content-Type": "text/csv"}
    fmt = "CSV"

    spec = spec_str.format(
        right_ascension,
        declination,
        epoch,
        search_radius,
        fmt,
        catalog
    )
    service_url = f"{SERVICELOCATION}/{service_type}?{spec}"
    try:
        rawcat = requests.get(service_url, headers=headers, timeout=timeout)
    except requests.exceptions.ConnectionError:
        raise requests.exceptions.ConnectionError(
            "Could not connect to the VO API server. Try again later."
        )
    except requests.exceptions.Timeout:
        raise requests.exceptions.Timeout("The request to the VO API server timed out.")
    except requests.exceptions.RequestException:
        raise requests.exceptions.RequestException(
            "There was an unexpected error with the request."
        )
    r_contents = rawcat.content.decode()  # convert from bytes to a String
    rstr = r_contents.split("\r\n")
    # remove initial line describing the number of sources returned
    # CRITICAL to proper interpretation of CSV data
    del rstr[0]

    return Table.read(rstr, format="csv")



import os
from pathlib import Path

import numpy as np
import pytest

from stcal.outlier_detection.median import (
    MedianComputer,
    _DiskAppendableArray,
    _OnDiskMedian,
    nanmedian3D,
)
from stcal.testing_helpers import MemoryThreshold


def test_disk_appendable_array(tmp_path):

    slice_shape = (8, 7)
    dtype = "float32"
    tempdir = tmp_path / Path("tmptest")
    Path.mkdir(tempdir)
    fname = tempdir / "test.bin"

    arr = _DiskAppendableArray(slice_shape, dtype, fname)

    # check temporary file setup
    assert str(arr._filename).split("/")[-1] in os.listdir(tempdir)  # noqa: SLF001
    assert len(os.listdir(tempdir)) == 1
    assert arr.shape == (0, *slice_shape)

    # check cwd contains no files
    assert all(not Path.is_file(Path(f)) for f in os.listdir(tmp_path))

    # check expected append failures
    candidate = np.zeros((7, 7), dtype=dtype)
    with pytest.raises(ValueError, match="shape"):
        arr.append(candidate)
    candidate = np.zeros((8, 7), dtype="float64")
    with pytest.raises(ValueError, match="dtype"):
        arr.append(candidate)

    # check append and read
    candidate0 = np.zeros(slice_shape, dtype=dtype)
    candidate1 = np.full(slice_shape, 2, dtype=dtype)
    candidate2 = np.full(slice_shape, np.nan, dtype=dtype)
    for candidate in [candidate0, candidate1, candidate2]:
        arr.append(candidate)

    arr_in_memory = arr.read()

    assert arr_in_memory.shape == (3, *slice_shape)
    assert np.all(arr_in_memory[0] == candidate0)
    assert np.all(arr_in_memory[1] == candidate1)
    assert np.allclose(arr_in_memory[2], candidate2, equal_nan=True)


def test_disk_appendable_array_bad_inputs(tmp_path):

    slice_shape = (8, 7)
    dtype = "float32"
    tempdir = tmp_path / Path("tmptest")
    fname = "test.bin"

    # test input directory does not exist
    with pytest.raises(FileNotFoundError):
        _DiskAppendableArray(slice_shape, dtype, tempdir / fname)

    # make the input directory
    Path.mkdir(tempdir)

    # ensure failure if slice_shape is not 2-D
    with pytest.raises(ValueError, match="slice shape"):
        _DiskAppendableArray((3, 5, 7), dtype, tempdir / fname)

    # ensure failure if dtype is not valid
    with pytest.raises(TypeError):
        _DiskAppendableArray(slice_shape, "float3", tempdir / fname)

    # ensure failure if pass directory instead of filename
    with pytest.raises(IsADirectoryError):
        _DiskAppendableArray(slice_shape, "float3", tempdir)


def test_on_disk_median(tmp_path):

    library_length = 3
    frame_shape = (21, 20)
    dtype = "float32"
    tempdir = tmp_path / Path("tmptest")
    Path.mkdir(tempdir)
    shape = (library_length, *frame_shape)

    median_computer = _OnDiskMedian(shape, dtype=dtype, tempdir=tempdir)

    # test compute buffer indices
    # buffer size equals size of single input model by default
    # which means we expect same number of sections as library length
    # in reality there is often one more section than that because
    # need integer number of rows per section, but math is exact in this case
    expected_buffer_size = frame_shape[0] * frame_shape[1] * \
        np.dtype(dtype).itemsize
    expected_section_nrows = frame_shape[0] // library_length
    assert median_computer.nsections == library_length
    assert median_computer.section_nrows == expected_section_nrows
    assert median_computer.buffer_size == expected_buffer_size

    # test temp file setup
    assert len(os.listdir(tempdir)) == 1
    assert str(median_computer._temp_path)\
        .startswith(str(tempdir))  # noqa: SLF001
    assert len(os.listdir(median_computer._temp_path)) \
        == library_length  # noqa: SLF001
    # check cwd and parent tempdir contain no files
    assert all(not Path.is_file(Path(f)) for f in os.listdir(tmp_path))
    assert all(not Path.is_file(Path(f)) for f in os.listdir(tempdir))

    # test validate data
    candidate = np.zeros((20, 20), dtype=dtype)
    with pytest.raises(ValueError, match="shape"):
        median_computer.add_image(candidate)
    candidate = np.zeros((21, 20), dtype="float64")
    with pytest.raises(ValueError, match="dtype"):
        median_computer.add_image(candidate)

    # test add and compute
    candidate0 = np.full(frame_shape, 3, dtype=dtype)
    candidate1 = np.full(frame_shape, 2, dtype=dtype)
    candidate2 = np.full(frame_shape, np.nan, dtype=dtype)
    for candidate in [candidate0, candidate1, candidate2]:
        median_computer.add_image(candidate)
    median = median_computer.compute_median()
    median_computer.cleanup()
    assert median.shape == frame_shape
    assert np.all(median == 2.5)

    # test expected error trying to add too many frames
    # for loop to ensure always happens, not just the first time
    candidate3 = np.zeros_like(candidate0)
    for _ in range(2):
        with pytest.raises(IndexError):
            median_computer.add_image(candidate3)

    # test cleanup of tmpdir and everything else
    assert not Path.exists(median_computer._temp_path)  # noqa: SLF001
    assert len(os.listdir(tempdir)) == 0


def test_computer():
    """Ensure MedianComputer works the same on disk and in memory"""
    full_shape = (3, 21, 20)
    comp_memory = MedianComputer(full_shape, True)
    comp_disk = MedianComputer(full_shape, False)
    for i in range(full_shape[0]):
        frame = np.full((21, 20), i, dtype=np.float32)
        comp_memory.append(frame, i)
        comp_disk.append(frame, i)
    assert np.allclose(comp_memory.evaluate(), comp_disk.evaluate())


def test_on_disk_median_bad_inputs(tmp_path):

    library_length = 3
    frame_shape = (21, 20)
    dtype = "float32"
    tempdir = tmp_path / Path("tmptest")
    Path.mkdir(tempdir)
    shape = (library_length, *frame_shape)

    with pytest.raises(ValueError, match="shape"):
        _OnDiskMedian(frame_shape, dtype=dtype, tempdir=tempdir)

    with pytest.raises(TypeError):
        _OnDiskMedian(shape, dtype="float3", tempdir=tempdir)

    with pytest.raises(FileNotFoundError):
        _OnDiskMedian(shape, dtype="float32", tempdir="dne")

    # ensure unreasonable buffer size will get set to minimum reasonable buffer
    min_buffer = np.dtype(dtype).itemsize*frame_shape[1]*library_length
    median_computer = _OnDiskMedian(shape,
                                   dtype=dtype,
                                   tempdir=tempdir,
                                   buffer_size=-1)
    assert median_computer.buffer_size == min_buffer
    median_computer.cleanup()


def test_nanmedian3D():

    shp = (11, 50, 60)
    generator = np.random.default_rng(77)
    cube = generator.normal(size=shp)
    cube[5, 5:7, 5:8] = np.nan
    med = nanmedian3D(cube.astype(np.float32))

    assert med.dtype == np.float32
    assert np.allclose(med, np.nanmedian(cube, axis=0), equal_nan=True)


@pytest.mark.parametrize("in_memory", [True, False])
def test_memory_computer(in_memory, tmp_path):
    """
    Analytically calculate how much memory the median computation
    is supposed to take, then ensure that the implementation
    stays near that.

    in_memory=True case allocates the following memory:
    - one cube size
    - median array == one frame size

    in_memory=False case allocates the following memory:
    - one buffer size, which by default is the frame size
    - median array == one frame size
    
    add a half-frame-size buffer to the expected memory usage in both cases
    """
    shp = (20, 500, 500)
    cube_size = np.dtype("float32").itemsize * shp[0] * shp[1] * shp[2] #bytes
    frame_size = cube_size / shp[0]

    # calculate expected memory usage
    if in_memory:
        expected_mem = cube_size + frame_size*1.5
    else:
        expected_mem = frame_size * 2.5

    # compute the median while tracking memory usage
    with MemoryThreshold(str(expected_mem) + " B"):
        computer = MedianComputer(shp, in_memory=in_memory, tempdir=tmp_path)
        for i in range(shp[0]):
            frame = np.full(shp[1:], i, dtype=np.float32)
            computer.append(frame, i)
            del frame
        computer.evaluate()



"""
Utility functions for outlier detection routines
"""
import warnings

import numpy as np
from astropy.stats import sigma_clip
from drizzle.cdrizzle import tblot
from scipy import ndimage
from skimage.util import view_as_windows
import gwcs

from stcal.alignment.util import wcs_bbox_from_shape

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


__all__ = [
    "medfilt",
    "compute_weight_threshold",
    "flag_crs",
    "flag_resampled_crs",
    "gwcs_blot",
    "calc_gwcs_pixmap",
    "reproject",
]


def medfilt(arr, kern_size):
    """
    scipy.signal.medfilt (and many other median filters) have undefined behavior
    for nan inputs. See: https://github.com/scipy/scipy/issues/4800

    Parameters
    ----------
    arr : numpy.ndarray
        The input array

    kern_size : list of int
        List of kernel dimensions, length must be equal to arr.ndim.

    Returns
    -------
    filtered_arr : numpy.ndarray
        Input array median filtered with a kernel of size kern_size
    """
    padded = np.pad(arr, [[k // 2] for k in kern_size])
    windows = view_as_windows(padded, kern_size, np.ones(len(kern_size), dtype='int'))
    return np.nanmedian(windows, axis=np.arange(-len(kern_size), 0))


def compute_weight_threshold(weight, maskpt):
    '''
    Compute the weight threshold for a single image or cube.

    Parameters
    ----------
    weight : numpy.ndarray
        The weight array

    maskpt : float
        The percentage of the mean weight to use as a threshold for masking.

    Returns
    -------
    float
        The weight threshold for this integration.
    '''
    return np.mean(
        sigma_clip(
            weight[np.isfinite(weight) & (weight != 0)],
            sigma=3,
            maxiters=5,
            masked=False,
            copy=False,
        ),
    dtype='f8') * maskpt


def _abs_deriv(array):
    """Take the absolute derivate of a numpy array."""
    out = np.zeros_like(array)  # use same dtype as input

    # make output values nan where input is nan (for floating point input)
    if np.issubdtype(array.dtype, np.floating):
        out[np.isnan(array)] = np.nan

    # compute row-wise absolute diffference
    row_diff = np.abs(np.diff(array, axis=0))
    np.putmask(out[1:], np.isfinite(row_diff), row_diff)  # no need to do max yet

    # since these are absolute differences |r0-r1| = |r1-r0|
    # make a view of the target portion of the array
    row_offset_view = out[:-1]
    # compute an in-place maximum
    np.putmask(row_offset_view, row_diff > row_offset_view, row_diff)
    del row_diff

    # compute col-wise absolute difference
    col_diff = np.abs(np.diff(array, axis=1))
    col_offset_view = out[:, 1:]
    np.putmask(col_offset_view, col_diff > col_offset_view, col_diff)
    col_offset_view = out[:, :-1]
    np.putmask(col_offset_view, col_diff > col_offset_view, col_diff)
    return out


def flag_crs(
    sci_data,
    sci_err,
    blot_data,
    snr,
):
    """
    Straightforward detection of outliers for non-dithered data since
    sci_err includes all noise sources (photon, read, and flat for baseline).

    Parameters
    ----------
    sci_data : numpy.ndarray
        "Science" data possibly containing outliers.

    sci_err : numpy.ndarray
        Error estimates for sci_data.

    blot_data : numpy.ndarray
        Reference data used to detect outliers.

    snr : float
        Signal-to-noise ratio used during detection.

    Returns
    -------
    cr_mask : numpy.ndarray
        Boolean array where outliers (CRs) are true.
    """
    return np.greater(np.abs(sci_data - blot_data), snr * np.nan_to_num(sci_err))


def flag_resampled_crs(
    sci_data,
    sci_err,
    blot_data,
    snr1,
    snr2,
    scale1,
    scale2,
    backg,
):
    """
    Detect outliers (CRs) using resampled reference data.

    Parameters
    ----------

    sci_data : numpy.ndarray
        "Science" data possibly containing outliers

    sci_err : numpy.ndarray
        Error estimates for sci_data

    blot_data : numpy.ndarray
        Reference data used to detect outliers.

    snr1 : float
        Signal-to-noise ratio threshold used prior to smoothing.

    snr2 : float
        Signal-to-noise ratio threshold used after smoothing.

    scale1 : float
        Scale used prior to smoothing.

    scale2 : float
        Scale used after smoothing.

    backg : float
        Scalar background to subtract from the difference.

    Returns
    -------
    cr_mask : numpy.ndarray
        boolean array where outliers (CRs) are true
    """
    err_data = np.nan_to_num(sci_err)

    blot_deriv = _abs_deriv(blot_data)
    diff_noise = np.abs(sci_data - blot_data - backg)

    # Create a boolean mask based on a scaled version of
    # the derivative image (dealing with interpolating issues?)
    # and the standard n*sigma above the noise
    threshold1 = scale1 * blot_deriv + snr1 * err_data
    mask1 = np.greater(diff_noise, threshold1)

    # Smooth the boolean mask with a 3x3 boxcar kernel
    kernel = np.ones((3, 3), dtype=int)
    mask1_smoothed = ndimage.convolve(mask1, kernel, mode='nearest')

    # Create a 2nd boolean mask based on the 2nd set of
    # scale and threshold values
    threshold2 = scale2 * blot_deriv + snr2 * err_data
    mask2 = np.greater(diff_noise, threshold2)

    # Final boolean mask
    return mask1_smoothed & mask2


def gwcs_blot(median_data, median_wcs, blot_shape, blot_wcs, pix_ratio, fillval=0.0):
    """
    Resample the median data to recreate an input image based on
    the blot wcs.

    Parameters
    ----------
    median_data : numpy.ndarray
        The data to blot.

    median_wcs : gwcs.wcs.WCS
        The wcs for the median data.

    blot_shape : tuple of int
        The target blot data shape.

    blot_wcs : gwcs.wcs.WCS
        The target/blotted wcs.

    pix_ratio : float
        Pixel ratio.

    fillval : float, optional
        Fill value for missing data.

    Returns
    -------
    blotted : numpy.ndarray
        The blotted median data.

    blot_img : datamodel
        Datamodel containing header and WCS to define the 'blotted' image
    """
    # Compute the mapping between the input and output pixel coordinates
    pixmap = calc_gwcs_pixmap(blot_wcs, median_wcs, blot_shape)
    log.debug("Pixmap shape: {}".format(pixmap[:, :, 0].shape))
    log.debug("Sci shape: {}".format(blot_shape))
    log.info('Blotting {} <-- {}'.format(blot_shape, median_data.shape))

    outsci = np.full(blot_shape, fillval, dtype=np.float32)

    # Currently tblot cannot handle nans in the pixmap, so we need to give some
    # other value.  -1 is not optimal and may have side effects.  But this is
    # what we've been doing up until now, so more investigation is needed
    # before a change is made.  Preferably, fix tblot in drizzle.
    pixmap[np.isnan(pixmap)] = -1
    tblot(median_data, pixmap, outsci, scale=pix_ratio, kscale=1.0,
          interp='linear', exptime=1.0, misval=fillval, sinscl=1.0)

    return outsci


def calc_gwcs_pixmap(in_wcs, out_wcs, in_shape):
    """
    Return a pixel grid map from input frame to output frame.

    Parameters
    ----------
    in_wcs : gwcs.wcs.WCS
        Input/source wcs.

    out_wcs : gwcs.wcs.WCS
        Output/projected wcs.

    in_shape : list of int
        Input shape used to compute the input bounding box.

    Returns
    -------
    pixmap : numpy.ndarray
        Computed pixmap.
    """
    bb = wcs_bbox_from_shape(in_shape)
    log.debug("Bounding box from data shape: {}".format(bb))

    grid = gwcs.wcstools.grid_from_bounding_box(bb)
    return np.dstack(reproject(in_wcs, out_wcs)(grid[0], grid[1]))


def reproject(wcs1, wcs2):
    """
    Given two WCSs return a function which takes pixel
    coordinates in wcs1 and computes them in wcs2.

    It performs the forward transformation of ``wcs1`` followed by the
    inverse of ``wcs2``.

    Parameters
    ----------
    wcs1, wcs2 : gwcs.wcs.WCS
        WCS objects that have `pixel_to_world_values` and `world_to_pixel_values`
        methods.

    Returns
    -------
    _reproject :
        Function to compute the transformations.  It takes x, y
        positions in ``wcs1`` and returns x, y positions in ``wcs2``.
    """

    try:
        forward_transform = wcs1.pixel_to_world_values
        backward_transform = wcs2.world_to_pixel_values
    except AttributeError as err:
        raise TypeError("Input should be a WCS") from err

    def _reproject(x, y):
        sky = forward_transform(x, y)
        flat_sky = []
        for axis in sky:
            flat_sky.append(axis.flatten())
        det = backward_transform(*tuple(flat_sky))
        det_reshaped = []
        for axis in det:
            det_reshaped.append(axis.reshape(x.shape))
        return tuple(det_reshaped)
    return _reproject



# cython: language_level=3str

from libcpp cimport bool
from libcpp.vector cimport vector

from stcal.ramp_fitting.ols_cas22._ramp cimport RampFit, RampQueue


cpdef enum FixedOffsets:
    single_t_bar_diff
    double_t_bar_diff
    single_t_bar_diff_sqr
    double_t_bar_diff_sqr
    single_read_recip
    double_read_recip
    single_var_slope_val
    double_var_slope_val
    n_fixed_offsets


cpdef enum PixelOffsets:
    single_local_slope
    double_local_slope
    single_var_read_noise
    double_var_read_noise
    n_pixel_offsets


cpdef enum:
    JUMP_DET = 4

cdef struct Thresh:
    float intercept
    float constant


cdef struct JumpFits:
    RampFit average
    vector[int] jumps
    vector[RampFit] fits
    RampQueue index


cpdef float[:, :] fill_fixed_values(float[:, :] fixed,
                                    float[:] t_bar,
                                    float[:] tau,
                                    int[:] n_reads,
                                    int n_resultants)


cdef JumpFits fit_jumps(float[:] resultants,
                        int[:] dq,
                        float read_noise,
                        float[:] t_bar,
                        float[:] tau,
                        int[:] n_reads,
                        int n_resultants,
                        float[:, :] fixed,
                        float[:, :] pixel,
                        Thresh thresh,
                        bool use_jump,
                        bool include_diagnostic)
